#!/usr/bin/env python

# pylint: disable=redefined-outer-name,too-many-arguments,unused-import

"""CLI tests."""

import logging
import json
import os

from contextlib import contextmanager
from pathlib import Path
from time import time
from tempfile import NamedTemporaryFile

import certifi
import pytest

from docker_registry_client_async import DockerRegistryClientAsync, ImageName
from docker_sign_verify import FormattedSHA256
from pytest_docker_apache_fixtures import ApacheSecure
from pytest_gnupg_fixtures import GnuPGKeypair
from _pytest.logging import LogCaptureFixture
from _pytest.tmpdir import TempPathFactory

from oc_mirror.scripts.atomic import cli

LOGGER = logging.getLogger(__name__)


@contextmanager
def drca_cacerts(path: Path):
    """Context manager to globally define the DRCA CA trust store."""
    key = "DRCA_CACERTS"
    old = os.environ.get(key, None)

    tmpfile = NamedTemporaryFile()
    tmpfile.write(path.read_bytes())
    tmpfile.write(Path(certifi.where()).read_bytes())
    tmpfile.flush()

    os.environ[key] = tmpfile.name
    yield None
    if old is not None:
        os.environ[key] = old
    else:
        del os.environ[key]

    tmpfile.close()


@contextmanager
def drca_credentials_store(apache_secure: ApacheSecure):
    """
    Context manager to globally define the DRCA credentials store.

    Args:
        apache_secure: The secure apache from which to retrieve the credentials.

    Yields:
        The path to the DRCA credentials store.
    """
    key = "DRCA_CREDENTIALS_STORE"
    old = os.environ.get(key, None)

    auth = apache_secure.auth_header["Authorization"].split()[1]
    credentials = {"auths": {f"https://{apache_secure.endpoint}": {"auth": auth}}}
    tmpfile = NamedTemporaryFile()
    tmpfile.write(json.dumps(credentials).encode("utf-8"))
    tmpfile.flush()

    os.environ[key] = str(tmpfile.name)
    yield None
    if old is not None:
        os.environ[key] = old
    else:
        del os.environ[key]
    tmpfile.close()


@contextmanager
def gnupghome(path: Path):
    """Context manager to globally define the GnuPG home directory."""
    key = "GNUPGHOME"
    old = os.environ.get(key, None)
    os.environ[key] = str(path)
    yield None
    if old is not None:
        os.environ[key] = old
    else:
        del os.environ[key]


def test_empty_args(clirunner):
    """Test atomic CLI can be invoked."""
    for command in ["sign", "verify"]:
        result = clirunner.invoke(cli, [command], catch_exceptions=False)
        assert "Usage:" in result.stdout
        assert result.exit_code != 0


@pytest.mark.online
def test_sign_bad_keyid(
    caplog: LogCaptureFixture,
    gnupg_keypair: GnuPGKeypair,
    runner,
    tmp_path_factory: TempPathFactory,
):
    """Test atomic can handle invalid keyids."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)

    image_name = ImageName.parse(f"foo/bar:{__name__}")
    image_name.digest = FormattedSHA256.calculate(
        f"random_value: {time()}".encode(encoding="utf-8")
    )
    gnupg_home = tmp_path_factory.mktemp(__name__)
    with gnupghome(gnupg_home):
        result = runner.invoke(
            cli,
            args=[
                "sign",
                "--keyid",
                "invalidkeyid",
                "--keypass",
                gnupg_keypair.passphrase,
                str(image_name),
            ],
        )
        assert isinstance(result.exception, SystemExit)
        assert (
            f"Unable to create new signature for digest: {image_name.digest}"
            in caplog.text
        )
        assert "Signature stored." not in caplog.text
        assert f"/sha256={image_name.digest.sha256}/signature-" not in caplog.text


@pytest.mark.online
def test_verify_no_signatures(
    apache_secure: ApacheSecure,
    caplog: LogCaptureFixture,
    runner,
):
    """Test atomic can operate on images without existing signatures."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)

    image_name = ImageName.parse(f"foo/bar:{__name__}")
    image_name.digest = FormattedSHA256.calculate(
        f"random_value: {time()}".encode(encoding="utf-8")
    )
    with drca_credentials_store(apache_secure), drca_cacerts(apache_secure.cacerts):
        result = runner.invoke(
            cli,
            args=[
                f"--signature-store=https://{apache_secure.endpoint}",
                "verify",
                str(image_name),
            ],
        )
    assert isinstance(result.exception, SystemExit)
    assert "Unable to locate a valid signature" in caplog.text
    assert "Verified" not in caplog.text
    assert str(image_name) in caplog.text


@pytest.mark.online_modification
def test_verify_not_found(
    apache_secure: ApacheSecure,
    caplog: LogCaptureFixture,
    runner,
):
    """Test atomic can handle incorrect image names."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)

    image_name = ImageName.parse(f"{apache_secure.endpoint}/foo/bar:{__name__}")
    with drca_credentials_store(apache_secure), drca_cacerts(apache_secure.cacerts):
        result = runner.invoke(
            cli,
            args=[
                f"--signature-store=https://{apache_secure.endpoint}",
                "verify",
                str(image_name),
            ],
        )
        assert isinstance(result.exception, SystemExit)
        assert "404" in caplog.text
        assert "Not Found" in caplog.text
        assert str(image_name) in caplog.text


@pytest.mark.online
def test_verify_signed(
    apache_secure: ApacheSecure,
    caplog: LogCaptureFixture,
    gnupg_keypair: GnuPGKeypair,
    runner,
):
    """Test atomic can handle signed images."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)

    image_name = ImageName.parse(f"foo/bar:{__name__}")
    image_name.digest = FormattedSHA256.calculate(
        f"random_value: {time()}".encode(encoding="utf-8")
    )
    for signature_type in ["image-config", "manifest"]:
        with drca_credentials_store(apache_secure), drca_cacerts(
            apache_secure.cacerts
        ), gnupghome(gnupg_keypair.gnupg_home):
            result = runner.invoke(
                cli,
                args=[
                    f"--signature-store=https://{apache_secure.endpoint}",
                    f"--signature-type={signature_type}",
                    "sign",
                    "--keyid",
                    str(gnupg_keypair.keyid),
                    "--keypass",
                    gnupg_keypair.passphrase,
                    str(image_name),
                ],
            )
            assert not result.exception
            assert "Signature stored." in caplog.text
            assert (
                f"https://{apache_secure.endpoint}/sha256={image_name.digest.sha256}/signature-"
                in caplog.text
            )

            caplog.clear()

            result = runner.invoke(
                cli,
                args=[
                    f"--signature-store=https://{apache_secure.endpoint}",
                    f"--signature-type={signature_type}",
                    "verify",
                    str(image_name),
                ],
            )
            assert not result.exception
            assert (
                f"https://{apache_secure.endpoint}/sha256={image_name.digest.sha256}/signature-"
                in caplog.text
            )
            assert "Signature matches:" in caplog.text
            assert "Signature is compliant:" in caplog.text


@pytest.mark.online
def test_verify_unauthorized(
    apache_secure: ApacheSecure,
    caplog: LogCaptureFixture,
    runner,
):
    """Test atomic can handle incorrect credentials."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)

    image_name = ImageName.parse(f"foo/bar:{__name__}")
    image_name.digest = FormattedSHA256.calculate(
        f"random_value: {time()}".encode(encoding="utf-8")
    )

    # Note: Not using apache_secure_credentials().
    with drca_cacerts(apache_secure.cacerts):
        result = runner.invoke(
            cli,
            args=[
                f"--signature-store=https://{apache_secure.endpoint}",
                "verify",
                str(image_name),
            ],
        )
        assert isinstance(result.exception, SystemExit)
        assert "Unable to locate a valid signature!" in caplog.text
        assert str(image_name) in caplog.text



#!/usr/bin/env python

# pylint: disable=protected-access,redefined-outer-name

"""OpenShift release tests."""

import asyncio
import logging

from typing import Dict, List, Set

import pytest

from docker_registry_client_async import FormattedSHA256, ImageName
from docker_sign_verify import RegistryV2
from _pytest.logging import LogCaptureFixture
from pytest_docker_registry_fixtures import DockerRegistrySecure

from oc_mirror.ocrelease import (
    get_release_metadata,
    log_release_metadata,
    put_release,
    translate_release_metadata,
    TypingRegexSubstitution,
)

from .testutils import equal_if_unqualified, needs_credentials

pytestmark = [pytest.mark.asyncio]

LOGGER = logging.getLogger(__name__)

# Bug Fix: https://github.com/crashvb/docker-registry-client-async/issues/24
#
# Right now this is known to leave a nasty "Fatal error on SSL transport" error
# at the end of the test execution; however, without this we cannot test using
# a TLS-in-TLS proxy ...
setattr(asyncio.sslproto._SSLProtocolTransport, "_start_tls_compatible", True)


@pytest.mark.online
@pytest.mark.parametrize(
    "release,count_blobs,count_manifests,count_signatures,count_signature_stores,count_signing_keys,known_good_blobs,"
    "known_good_manifests,manifest_digest",
    [
        (
            "quay.io/openshift-release-dev/ocp-release:4.4.6-x86_64",
            227,
            109,
            2,
            2,
            1,
            {
                "sha256:06be4357dfb813c8d3d828b95661028d3d2a380ed8909b60c559770c0cd2f917": [
                    "quay.io/openshift-release-dev/ocp-release"
                ],
                "sha256:49be5ad10f908f0b5917ba11ab8529d432282fd6df7b8a443d60455619163b9c": [
                    "quay.io/openshift-release-dev/ocp-v4.0-art-dev"
                ],
            },
            {
                "quay.io/openshift-release-dev/ocp-release@sha256:7613d8f7db639147b91b16b54b24cfa351c3cbde6aa7b7bf1b9c8"
                "0c260efad06": "4.4.6-x86_64",
                "quay.io/openshift-release-dev/ocp-v4.0-art-dev@sha256:ce1f23618369fc00eab1f9a9bb5f409ed6a3c2652770c807"
                "7a099a69064ee436": "4.4.6-aws-machine-controllers",
            },
            "sha256:7613d8f7db639147b91b16b54b24cfa351c3cbde6aa7b7bf1b9c80c260efad06",
        )
    ],
)
@needs_credentials("quay.io")
async def test_get_release_metadata(
    registry_v2_proxy: RegistryV2,
    release: str,
    count_blobs: int,
    count_manifests: int,
    count_signatures: int,
    count_signature_stores: int,
    count_signing_keys: int,
    known_good_blobs: Dict[FormattedSHA256, Set[str]],
    known_good_manifests: Dict[ImageName, str],
    manifest_digest: FormattedSHA256,
):
    # pylint: disable=too-many-arguments
    """Tests release metadata retrieval from a remote registry."""
    logging.getLogger("gnupg").setLevel(logging.FATAL)

    # Retrieve the release metadata ...
    image_name = ImageName.parse(release)
    result = await get_release_metadata(
        registry_v2=registry_v2_proxy, release_name=image_name
    )

    assert result.blobs
    assert len(result.blobs) == count_blobs
    for digest in known_good_blobs.keys():
        assert digest in result.blobs.keys()
        for image_prefix in known_good_blobs[digest]:
            assert image_prefix in result.blobs[digest]

    assert result.manifest_digest
    assert result.manifest_digest == manifest_digest

    assert result.manifests
    assert len(result.manifests) == count_manifests
    for image_name in known_good_manifests.keys():
        assert result.manifests[image_name] == known_good_manifests[image_name]

    assert result.raw_image_references

    assert result.raw_release_metadata

    assert result.signatures
    assert len(result.signatures) == count_signatures

    assert result.signature_stores
    assert len(result.signature_stores) == count_signature_stores

    assert result.signing_keys
    assert len(result.signing_keys) == count_signing_keys


@pytest.mark.online
@pytest.mark.parametrize(
    "release,fingerprint,username",
    [
        (
            "quay.io/openshift-release-dev/ocp-release:4.4.6-x86_64",
            "567E347AD0044ADE55BA8A5F199E2F91FD431D51",
            "Red Hat, Inc. (release key 2) <security@redhat.com>",
        )
    ],
)
@needs_credentials("quay.io")
async def test_log_release_metadata(
    caplog: LogCaptureFixture,
    fingerprint: str,
    registry_v2_proxy: RegistryV2,
    release: str,
    username: str,
):
    """Tests logging of release metadata."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    logging.getLogger("gnupg").setLevel(logging.FATAL)

    # Retrieve the release metadata ...
    image_name = ImageName.parse(release)
    result = await get_release_metadata(
        registry_v2=registry_v2_proxy,
        release_name=image_name,
    )
    assert result

    await log_release_metadata(release_metadata=result, release_name=image_name)
    assert str(image_name) in caplog.text
    assert "quay.io/openshift-release-dev/ocp-v4.0-art-dev@sha256:" in caplog.text
    assert "atomic container signature" in caplog.text
    assert fingerprint in caplog.text
    assert username in caplog.text


@pytest.mark.online_modification
@pytest.mark.parametrize(
    "release", ["quay.io/openshift-release-dev/ocp-release:4.4.6-x86_64"]
)
@needs_credentials("quay.io")
async def test_put_release_from_internet(
    docker_registry_secure: DockerRegistrySecure,
    registry_v2_proxy: RegistryV2,
    release: str,
):
    """Tests release replication to a local registry."""
    logging.getLogger("gnupg").setLevel(logging.FATAL)

    # Retrieve the release metadata ...
    image_name_src = ImageName.parse(release)
    release_metadata_src = await get_release_metadata(
        registry_v2=registry_v2_proxy,
        release_name=image_name_src,
        verify=False,
    )

    # Replicate the release ...
    image_name_dest = image_name_src.clone()
    image_name_dest.endpoint = docker_registry_secure.endpoint_name
    await put_release(
        release_name=image_name_dest,
        registry_v2=registry_v2_proxy,
        release_metadata=release_metadata_src,
    )

    # Retrieve the release metadata (again) ...
    release_metadata_dest = await get_release_metadata(
        registry_v2=registry_v2_proxy,
        release_name=image_name_dest,
    )

    # Release metadata should have the same blob digests ...
    assert (
        list(release_metadata_dest.blobs.keys()).sort()
        == list(release_metadata_src.blobs.keys()).sort()
    )
    # ... all blobs should correspond to the same namespaces ...
    for digest in release_metadata_src.blobs.keys():
        assert (
            list(release_metadata_dest.blobs[digest]).sort()
            == list(release_metadata_src.blobs[digest]).sort()
        )

    # Release metadata digest should be the same ...
    assert release_metadata_dest.manifest_digest == release_metadata_src.manifest_digest

    # Release metadata manifest digest should be the same ...
    assert (
        list(release_metadata_dest.manifests.keys()).sort()
        == list(release_metadata_src.manifests.keys()).sort()
    )

    # Translate the release image tags to a digest for comparison ...
    image_name_dest_digest = image_name_dest.clone()
    image_name_dest_digest.digest = release_metadata_dest.manifest_digest
    image_name_dest_digest.tag = None
    image_name_src_digest = image_name_src.clone()
    image_name_src_digest.digest = release_metadata_src.manifest_digest
    image_name_src_digest.tag = None

    # Release metadata manifest tags should be the same ...
    for image_name in release_metadata_src.manifests.keys():
        # Special Case: The release image in imposed in the metadata, not derived ...
        if equal_if_unqualified(image_name, image_name_src_digest):
            assert (
                release_metadata_dest.manifests[image_name_dest_digest]
                == release_metadata_src.manifests[image_name_src_digest]
            )
        else:
            assert (
                release_metadata_dest.manifests[image_name]
                == release_metadata_src.manifests[image_name]
            )

    # The raw image references should be the same ...
    assert (
        release_metadata_dest.raw_image_references.get_digest()
        == release_metadata_src.raw_image_references.get_digest()
    )

    # The raw release metadata should be the same ...
    assert (
        release_metadata_dest.raw_release_metadata
        == release_metadata_src.raw_release_metadata
    )

    # TODO: Do we need to check signatures here?

    # The signature stores should be the same ...
    assert (
        release_metadata_dest.signature_stores.sort()
        == release_metadata_src.signature_stores.sort()
    )

    # The signing keys should be the same ...
    assert (
        release_metadata_dest.signing_keys.sort()
        == release_metadata_src.signing_keys.sort()
    )


@pytest.mark.online_modification
@pytest.mark.parametrize(
    "release", ["quay.io/openshift-release-dev/ocp-release:4.4.6-x86_64"]
)
@needs_credentials("quay.io")
async def test_put_release_from_internal(
    docker_registry_secure_list: List[DockerRegistrySecure],
    registry_v2_list_proxy: RegistryV2,
    release: str,
):
    # pylint: disable=too-many-locals
    """Tests release replication to a local registry."""
    logging.getLogger("gnupg").setLevel(logging.FATAL)

    # Retrieve the release metadata (hop 0)...
    image_name0 = ImageName.parse(release)
    release_metadata0 = await get_release_metadata(
        registry_v2=registry_v2_list_proxy,
        release_name=image_name0,
        verify=False,
    )

    # Replicate the release (hop 1)...
    image_name1 = image_name0.clone()
    image_name1.endpoint = docker_registry_secure_list[0].endpoint_name
    await put_release(
        release_name=image_name1,
        registry_v2=registry_v2_list_proxy,
        release_metadata=release_metadata0,
        verify=False,
    )

    # Retrieve the release metadata (hop 1) ...
    release_metadata1 = await get_release_metadata(
        registry_v2=registry_v2_list_proxy,
        release_name=image_name1,
        verify=False,
    )

    # Translate to the second registry ...
    regex_substitutions = [
        TypingRegexSubstitution(
            pattern=r"quay\.io",
            replacement=docker_registry_secure_list[0].endpoint_name,
        )
    ]
    release_metadata1_translated = await translate_release_metadata(
        regex_substitutions=regex_substitutions, release_metadata=release_metadata1
    )

    # Replicate the release (hop 2) ...
    image_name2 = image_name0.clone()
    image_name2.endpoint = docker_registry_secure_list[1].endpoint_name
    await put_release(
        release_name=image_name2,
        registry_v2=registry_v2_list_proxy,
        release_metadata=release_metadata1_translated,
    )

    # Retrieve the release metadata (hop 2) ...
    release_metadata2 = await get_release_metadata(
        registry_v2=registry_v2_list_proxy,
        release_name=image_name2,
    )

    # Release metadata should have the same blob digests ...
    assert (
        list(release_metadata2.blobs.keys()).sort()
        == list(release_metadata0.blobs.keys()).sort()
    )
    # ... all blobs should correspond to the same namespaces ...
    for digest in release_metadata0.blobs.keys():
        assert (
            list(release_metadata2.blobs[digest]).sort()
            == list(release_metadata0.blobs[digest]).sort()
        )

    # Release metadata digest should be the same ...
    assert release_metadata2.manifest_digest == release_metadata0.manifest_digest

    # Release metadata manifest digest should be the same ...
    assert (
        list(release_metadata2.manifests.keys()).sort()
        == list(release_metadata0.manifests.keys()).sort()
    )

    # Translate the release image tags to a digest for comparison ...
    image_name0_digest = image_name0.clone()
    image_name0_digest.digest = release_metadata0.manifest_digest
    image_name0_digest.tag = None
    image_name2_digest = image_name2.clone()
    image_name2_digest.digest = release_metadata2.manifest_digest
    image_name2_digest.tag = None

    # Release metadata manifest tags should be the same ...
    for image_name in release_metadata0.manifests.keys():
        # Special Case: The release image in imposed in the metadata, not derived ...
        if equal_if_unqualified(image_name, image_name0_digest):
            assert (
                release_metadata2.manifests[image_name2_digest]
                == release_metadata0.manifests[image_name0_digest]
            )
        else:
            assert (
                release_metadata2.manifests[image_name]
                == release_metadata0.manifests[image_name]
            )

    # The raw image references should be the same ...
    assert (
        release_metadata2.raw_image_references.get_digest()
        == release_metadata0.raw_image_references.get_digest()
    )

    # The raw release metadata should be the same ...
    assert (
        release_metadata2.raw_release_metadata == release_metadata0.raw_release_metadata
    )

    # TODO: Do we need to check signatures here?

    # The signature stores should be the same ...
    assert (
        release_metadata2.signature_stores.sort()
        == release_metadata0.signature_stores.sort()
    )

    # The signing keys should be the same ...
    assert (
        release_metadata2.signing_keys.sort() == release_metadata0.signing_keys.sort()
    )



#!/usr/bin/env python

"""Operator mirror command line interface."""

import logging
import sys

from pathlib import Path
from traceback import print_exception
from typing import Dict, List, NamedTuple, Optional

import click

from click.core import Context
from docker_registry_client_async import ImageName
from docker_sign_verify import RegistryV2
from docker_sign_verify.scripts.utils import (
    async_command,
    LOGGING_DEFAULT,
    logging_options,
    set_log_levels,
    to_image_name,
)

from oc_mirror.oprelease import (
    get_release_metadata,
    log_release_metadata,
    put_release,
    translate_release_metadata,
    TypingGetReleaseMetadata,
    TypingRegexSubstitution,
)

from .utils import OPENSHIFT_SIGNATURE_STORES, version

LOGGER = logging.getLogger(__name__)


class TypingContextObject(NamedTuple):
    # pylint: disable=missing-class-docstring
    check_signatures: bool
    registry_v2: RegistryV2
    signature_stores: List[str]
    signing_keys: List[str]
    verbosity: int


def _convert_package_channel(*, package_channel: Optional[List[str]]) -> Dict[str, str]:
    """Convert tuple of ":" separated pairs into a dictionary, or None."""
    if package_channel:
        package_channel = [
            item.split(":") if ":" in item else (item, None) for item in package_channel
        ]
        package_channel = {package: channel for package, channel in package_channel}
    return package_channel


def get_context_object(context: Context) -> TypingContextObject:
    """Wrapper method to enforce type checking."""
    return context.obj


@click.group()
@click.option(
    "--check-signatures/--no-check-signatures",
    default=True,
    help="Toggles integrity vs integrity and signature checking.",
    show_default=True,
)
@click.option(
    "--dry-run", help="Do not write to destination image sources.", is_flag=True
)
@click.option(
    "-s",
    "--signature-store",
    envvar="OPM_SIGNATURE_STORE",
    help="Url of a signature store to use for retrieving signatures. Can be passed multiple times.",
    multiple=True,
)
@click.option(
    "-k",
    "--signing-key",
    envvar="OPM_SIGNING_KEY",
    help="Armored GnuPG trust store to use for signature verification. Can be passed multiple times.",
    multiple=True,
    type=click.Path(exists=True, dir_okay=False),
)
@logging_options
@click.pass_context
def cli(
    context: Context,
    check_signatures: bool,
    dry_run: False,
    signature_store: List[str],
    signing_key: List[str],
    verbosity: int = LOGGING_DEFAULT,
):
    # pylint: disable=too-many-arguments
    """Utilities for working with operator releases."""
    if verbosity is None:
        verbosity = LOGGING_DEFAULT

    set_log_levels(verbosity)

    if not signature_store:
        signature_store = OPENSHIFT_SIGNATURE_STORES

    signing_keys = []
    for path in [Path(x) for x in signing_key]:
        LOGGER.debug("Loading signing key: %s", path)
        signing_keys.append(path.read_text("utf-8"))

    context.obj = TypingContextObject(
        check_signatures=check_signatures,
        registry_v2=RegistryV2(dry_run=dry_run),
        signature_stores=signature_store,
        signing_keys=signing_keys,
        verbosity=verbosity,
    )


@cli.command()
@click.argument("index_name", callback=to_image_name, required=True)
@click.argument("package_channel", nargs=-1)
@click.option("--sort-metadata", help="Sort metadata keys.", is_flag=True)
@click.option(
    "--translate",
    help="Translate the registry endpoint(s) based on the index location.",
    is_flag=True,
)
@click.pass_context
@async_command
async def dump(
    context: Context,
    index_name: ImageName,
    package_channel: List[str],
    sort_metadata: bool = False,
    translate: bool = False,
) -> TypingGetReleaseMetadata:
    """Dumps the metadata for an operator release(s)."""
    ctx = get_context_object(context)

    try:
        LOGGER.info("Retrieving metadata for index: %s ...", index_name)
        release_metadata = await get_release_metadata(
            index_name=index_name,
            package_channel=_convert_package_channel(package_channel=package_channel),
            registry_v2=ctx.registry_v2,
            signature_stores=ctx.signature_stores,
            signing_keys=ctx.signing_keys,
            verify=ctx.check_signatures,
        )
        if translate:
            regex_substitutions = [
                TypingRegexSubstitution(
                    pattern=r"quay\.io", replacement=index_name.endpoint
                ),
                TypingRegexSubstitution(
                    pattern=r"registry\.redhat\.io", replacement=index_name.endpoint
                ),
            ]
            release_metadata = await translate_release_metadata(
                regex_substitutions=regex_substitutions,
                release_metadata=release_metadata,
            )
        await log_release_metadata(
            index_name=index_name,
            release_metadata=release_metadata,
            sort_metadata=sort_metadata,
        )
        return release_metadata
    except Exception as exception:  # pylint: disable=broad-except
        if ctx.verbosity > 0:
            logging.fatal(exception)
        if ctx.verbosity > LOGGING_DEFAULT:
            exc_info = sys.exc_info()
            print_exception(*exc_info)
        sys.exit(1)
    finally:
        await ctx.registry_v2.close()


@cli.command()
@click.argument("index_name_src", callback=to_image_name, required=True)
@click.argument("index_name_dest", callback=to_image_name, required=True)
@click.argument("package_channel", nargs=-1)
@click.pass_context
@async_command
async def mirror(
    context: Context,
    index_name_dest: ImageName,
    index_name_src: ImageName,
    package_channel: List[str],
):
    """Replicates a operator release(s) between a source and destination registry(ies)."""
    ctx = get_context_object(context)
    try:
        LOGGER.info("Retrieving metadata for index: %s ...", index_name_src)
        release_metadata = await get_release_metadata(
            index_name=index_name_src,
            package_channel=_convert_package_channel(package_channel=package_channel),
            registry_v2=ctx.registry_v2,
            signature_stores=ctx.signature_stores,
            signing_keys=ctx.signing_keys,
            verify=ctx.check_signatures,
        )
        regex_substitutions = [
            TypingRegexSubstitution(
                pattern=r"quay\.io", replacement=index_name_src.endpoint
            ),
            TypingRegexSubstitution(
                pattern=r"registry\.redhat\.io", replacement=index_name_src.endpoint
            ),
        ]
        release_metadata_translated = await translate_release_metadata(
            regex_substitutions=regex_substitutions,
            release_metadata=release_metadata,
        )
        LOGGER.info("Mirroring index to: %s ...", index_name_dest)
        await put_release(
            index_name=index_name_dest,
            registry_v2=ctx.registry_v2,
            release_metadata=release_metadata_translated,
            verify=False,  # Already verified above (or not =/) ...
        )
        if ctx.registry_v2.dry_run:
            LOGGER.info(
                "Dry run completed for index: %s", index_name_dest.resolve_name()
            )
        else:
            LOGGER.info("Mirrored index to: %s", index_name_dest.resolve_name())
    except Exception as exception:  # pylint: disable=broad-except
        if ctx.verbosity > 0:
            logging.fatal(exception)
        if ctx.verbosity > LOGGING_DEFAULT:
            exc_info = sys.exc_info()
            print_exception(*exc_info)
        sys.exit(1)
    finally:
        await ctx.registry_v2.close()


cli.add_command(version)

if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    cli()



#!/usr/bin/env python

# pylint: disable=redefined-outer-name

"""GPGSigner tests."""

import logging

from time import time
from typing import Generator

import pytest

from docker_registry_client_async import (
    DockerRegistryClientAsync,
    FormattedSHA256,
    ImageName,
)
from docker_sign_verify import GPGTrust
from pytest_docker_apache_fixtures import ApacheSecure
from pytest_gnupg_fixtures import GnuPGKeypair

from oc_mirror import AtomicSigner

pytestmark = [pytest.mark.asyncio]

LOGGER = logging.getLogger(__name__)


@pytest.fixture
async def docker_registry_client_async(
    apache_secure: ApacheSecure,
) -> Generator[DockerRegistryClientAsync, None, None]:
    """Provides a DockerRegistryClientAsync instance."""
    # Do not use caching; get a new instance for each test
    async with DockerRegistryClientAsync(
        ssl=apache_secure.ssl_context
    ) as docker_registry_client_async:
        credentials = apache_secure.auth_header["Authorization"].split()[1]
        await docker_registry_client_async.add_credentials(
            f"https://{apache_secure.endpoint}", credentials
        )
        yield docker_registry_client_async


@pytest.fixture()
def signaturestore(apache_secure: ApacheSecure) -> str:
    """Provides a modifiable signature store location."""
    return f"https://{apache_secure.endpoint}"


@pytest.fixture()
def atomicsigner(
    docker_registry_client_async: DockerRegistryClientAsync,
    gnupg_keypair: GnuPGKeypair,
    signaturestore: str,
) -> AtomicSigner:
    """Provides AtomicSigner instances."""
    return AtomicSigner(
        docker_registry_client_async=docker_registry_client_async,
        homedir=gnupg_keypair.gnupg_home,
        keyid=gnupg_keypair.fingerprints[1],
        locations=[signaturestore],
        passphrase=gnupg_keypair.passphrase,
    )


# TODO: Add tests for protected methods ...


async def test_simple(atomicsigner: AtomicSigner, gnupg_keypair: GnuPGKeypair):
    """Test configuration signing and verification using GPG."""
    digest = FormattedSHA256.calculate(f"TEST DATA: {time()}".encode(encoding="utf-8"))
    image_name = ImageName.parse(f"foo/bar:{time()}")
    LOGGER.debug("Test Data:\n  digest     : %s\n  image name : %s", digest, image_name)

    # Generate a signature for the test data ...
    find_all_signatures = await atomicsigner.atomicsign(
        digest=digest, image_name=image_name
    )
    assert find_all_signatures
    assert find_all_signatures.index
    assert find_all_signatures.signature
    assert find_all_signatures.url
    assert all(
        (
            value in find_all_signatures.url
            for value in ["https", digest.replace(":", "="), "signature-"]
        )
    )

    # Verify the generated signature against the test data ...
    results = await atomicsigner.atomicverify(digest=digest, image_name=image_name)
    assert results
    for result in results:
        assert result.fingerprint == atomicsigner.keyid
        assert atomicsigner.keyid.endswith(result.key_id)
        assert "failed" not in result.signer_long
        assert "failed" not in result.signer_short
        assert result.status_atomic is None
        assert result.status_gpg == "signature valid"
        assert result.timestamp
        assert result.trust == GPGTrust.ULTIMATE.value
        assert result.type == "atomicsigner"
        assert result.username == gnupg_keypair.uids[0]
        assert result.valid


@pytest.mark.skip("Test scenario not defined.")
async def test_bad_data(atomicsigner: AtomicSigner, gnupg_keypair: GnuPGKeypair):
    """Test configuration signing and verification using GPG with bad data."""
    digest = FormattedSHA256.calculate(f"TEST DATA: {time()}".encode(encoding="utf-8"))
    image_name = ImageName.parse(f"foo/bar:{time()}")
    LOGGER.debug("Test Data:\n  digest     : %s\n  image name : %s", digest, image_name)

    # Generate a signature for the test data ...
    find_all_signatures = await atomicsigner.atomicsign(
        digest=digest, image_name=image_name
    )
    assert find_all_signatures
    assert find_all_signatures.index
    assert find_all_signatures.signature
    assert find_all_signatures.url
    assert all(
        (
            value in find_all_signatures.url
            for value in ["https", digest.replace(":", "="), "signature-"]
        )
    )

    # TODO: How to tamper with the data (e.g. produce and stage a signature with a mismatched digest)?

    # Verify the generated signature against the test data ...
    results = await atomicsigner.atomicverify(digest=digest, image_name=image_name)
    assert results
    for result in results:
        assert result.fingerprint is None
        assert atomicsigner.keyid.endswith(result.key_id)
        assert "failed" not in result.signer_long
        assert "failed" not in result.signer_short
        assert result.status_atomic is not None
        assert result.status_gpg != "signature valid"
        assert result.timestamp is None
        assert result.trust == GPGTrust.UNDEFINED.value
        assert result.type == "gpg"
        assert result.username == gnupg_keypair.uids[0]
        assert not result.valid



#!/usr/bin/env python

# pylint: disable=redefined-outer-name

"""Configures execution of pytest."""

from ssl import create_default_context
from typing import Generator, List

import certifi
import pytest

from aiohttp.helpers import BasicAuth
from click.testing import CliRunner
from docker_sign_verify import RegistryV2
from pytest_asyncio.plugin import Mode
from pytest_docker_registry_fixtures import DockerRegistrySecure
from pytest_docker_squid_fixtures import SquidSecure

from .testutils import OCMCliRunner


def pytest_addoption(parser):
    """pytest add option."""
    parser.addoption(
        "--allow-online",
        action="store_true",
        default=False,
        help="Allow execution of online tests.",
    )
    parser.addoption(
        "--allow-online-modification",
        action="store_true",
        default=False,
        help="Allow modification of online content (implies --allow-online).",
    )


def pytest_collection_modifyitems(config, items):
    """pytest collection modifier."""

    skip_online = pytest.mark.skip(
        reason="Execution of online tests requires --allow-online option."
    )
    skip_online_modification = pytest.mark.skip(
        reason="Modification of online content requires --allow-online-modification option."
    )
    for item in items:
        if "online_modification" in item.keywords and not config.getoption(
            "--allow-online-modification"
        ):
            item.add_marker(skip_online_modification)
        elif (
            "online" in item.keywords
            and not config.getoption("--allow-online")
            and not config.getoption("--allow-online-modification")
        ):
            item.add_marker(skip_online)


def pytest_configure(config):
    """pytest configuration hook."""
    config.addinivalue_line("markers", "online: allow execution of online tests.")
    config.addinivalue_line(
        "markers", "online_modification: allow modification of online content."
    )

    config.option.asyncio_mode = Mode.AUTO


@pytest.fixture(scope="session")
def pdrf_scale_factor() -> int:
    """Scale PDRF to 2."""
    return 2


@pytest.fixture(scope="session")
def pdsf_scale_factor() -> int:
    """Scale PDSF to 2."""
    return 2


@pytest.fixture
def clirunner() -> Generator[CliRunner, None, None]:
    """Provides a runner for testing click command line interfaces."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        yield runner


# TODO: What is the best way to code `DRCA_DEBUG=1 DRCA_CREDENTIALS_STORE=~/.docker/quay.io-pull-secret.json` into
#       this fixture?
@pytest.fixture
async def registry_v2(
    docker_registry_secure: DockerRegistrySecure,
) -> RegistryV2:
    """Provides a RegistryV2 instance."""
    # Do not use caching; get a new instance for each test
    ssl_context = docker_registry_secure.ssl_context
    ssl_context.load_verify_locations(cafile=certifi.where())
    async with RegistryV2(ssl=ssl_context) as registry_v2:
        credentials = docker_registry_secure.auth_header["Authorization"].split()[1]
        await registry_v2.docker_registry_client_async.add_credentials(
            docker_registry_secure.endpoint, credentials
        )

        yield registry_v2


@pytest.fixture
async def registry_v2_proxy(
    docker_registry_secure: DockerRegistrySecure, squid_secure: SquidSecure
) -> RegistryV2:
    """Provides a RegistryV2 instance."""
    ssl_context = create_default_context(
        cadata=squid_secure.certs.ca_certificate.read_text("utf-8")
        + docker_registry_secure.certs.ca_certificate.read_text("utf-8")
    )
    ssl_context.load_verify_locations(cafile=certifi.where())
    # Do not use caching; get a new instance for each test
    async with RegistryV2(ssl=ssl_context) as registry_v2_proxy:
        credentials = docker_registry_secure.auth_header["Authorization"].split()[1]
        for name in [
            docker_registry_secure.endpoint,
            docker_registry_secure.endpoint_name,
        ]:
            await registry_v2_proxy.docker_registry_client_async.add_credentials(
                name, credentials
            )
        registry_v2_proxy.docker_registry_client_async.proxies[
            "https"
        ] = f"https://{squid_secure.endpoint}/"
        registry_v2_proxy.docker_registry_client_async.proxy_auth = BasicAuth(
            login=squid_secure.username, password=squid_secure.password
        )

        yield registry_v2_proxy


@pytest.fixture
async def registry_v2_list(
    docker_registry_secure_list: List[DockerRegistrySecure],
) -> RegistryV2:
    """Provides a RegistryV2 instance."""
    # Do not use caching; get a new instance for each test
    ssl_context = create_default_context(cafile=certifi.where())
    for docker_registry_secure in docker_registry_secure_list:
        ssl_context.load_verify_locations(cafile=str(docker_registry_secure.cacerts))
    async with RegistryV2(ssl=ssl_context) as registry_v2:
        for docker_registry_secure in docker_registry_secure_list:
            credentials = docker_registry_secure.auth_header["Authorization"].split()[1]
            await registry_v2.docker_registry_client_async.add_credentials(
                docker_registry_secure.endpoint, credentials
            )

        yield registry_v2


@pytest.fixture
async def registry_v2_list_proxy(
    docker_registry_secure_list: List[DockerRegistrySecure],
    squid_secure_list: List[SquidSecure],
) -> RegistryV2:
    """Provides a RegistryV2 instance."""
    ssl_context = create_default_context(cafile=certifi.where())
    for docker_registry_secure in docker_registry_secure_list:
        ssl_context.load_verify_locations(cafile=str(docker_registry_secure.cacerts))
    for squid_secure in squid_secure_list:
        ssl_context.load_verify_locations(cafile=str(squid_secure.cacerts))
    # Do not use caching; get a new instance for each test
    async with RegistryV2(ssl=ssl_context) as registry_v2_proxy:
        for docker_registry_secure in docker_registry_secure_list:
            credentials = docker_registry_secure.auth_header["Authorization"].split()[1]
            for name in [
                docker_registry_secure.endpoint,
                docker_registry_secure.endpoint_name,
            ]:
                await registry_v2_proxy.docker_registry_client_async.add_credentials(
                    name, credentials
                )
        for squid_secure in squid_secure_list:
            registry_v2_proxy.docker_registry_client_async.proxies[
                "https"
            ] = f"https://{squid_secure.endpoint}/"
            registry_v2_proxy.docker_registry_client_async.proxy_auth = BasicAuth(
                login=squid_secure.username, password=squid_secure.password
            )

        yield registry_v2_proxy


@pytest.fixture
def runner() -> Generator[OCMCliRunner, None, None]:
    """Provides a runner for testing click command line interfaces."""
    runner = OCMCliRunner()
    with runner.isolated_filesystem():
        yield runner



#!/usr/bin/env python

"""Utility classes."""

import logging
import tarfile

from io import BytesIO
from tempfile import TemporaryDirectory
from typing import List, NamedTuple

from aiotempfile.aiotempfile import open as aiotempfile
from docker_registry_client_async import FormattedSHA256, ImageName, Manifest
from docker_registry_client_async.utils import must_be_equal
from docker_sign_verify import (
    NoSignatureError,
    RegistryV2ManifestList,
    RegistryV2,
    SignatureMismatchError,
)
from docker_sign_verify.exceptions import DigestMismatchError
from docker_sign_verify.utils import be_kind_rewind, chunk_file
from pretty_bad_protocol import GPG
from pretty_bad_protocol._parsers import Verify
from pretty_bad_protocol._util import _make_binary_stream

from .atomicsignature import AtomicSignature
from .atomicsigner import AtomicSigner, AtomicSignerVerify

LOGGER = logging.getLogger(__name__)


class TypingDetachedSignature(NamedTuple):
    # pylint: disable=missing-class-docstring
    atomic_signature: AtomicSignature
    key: str
    raw_signature: bytes
    timestamp: str
    verify: Verify
    url: str
    username: str


class TypingRegexSubstitution(NamedTuple):
    # pylint: disable=missing-class-docstring
    pattern: str
    replacement: str


class TypingRetrieveAndVerifyReleaseMetadata(NamedTuple):
    # pylint: disable=missing-class-docstring
    signatures: List[TypingDetachedSignature]


class TypingVerifyDetachedSignature(NamedTuple):
    # pylint: disable=missing-class-docstring
    atomic_signature: AtomicSignature
    crypt: Verify
    result: bool


async def copy_blob(
    *,
    digest: FormattedSHA256,
    image_name_dest: ImageName,
    image_name_src: ImageName,
    registry_v2: RegistryV2,
):
    """
    Replicates a blob between two docker registry images.

    Args:
        digest: The digest of the blob to be replicated.
        image_name_dest: The destination registry image.
        image_name_src:  The source registry image.
        registry_v2: The underlying registry v2 image source to use to replicate.
    """
    docker_registry_client_async = registry_v2.docker_registry_client_async

    LOGGER.debug("Copying blob %s ...", digest)
    response = await docker_registry_client_async.head_blob(
        image_name=image_name_dest, digest=digest
    )
    if response.result:
        return

    async with aiotempfile(mode="w+b") as file:
        await docker_registry_client_async.get_blob_to_disk(
            digest=digest, file=file, image_name=image_name_src
        )
        await be_kind_rewind(file)
        await registry_v2.put_image_layer_from_disk(
            file=file, digest_expected=digest, image_name=image_name_dest
        )


async def copy_image(
    *,
    image_name_dest: ImageName,
    image_name_src: ImageName,
    registry_v2: RegistryV2,
):
    """
    Replicates an image between two docker registries.

    Args:
        image_name_dest: The destination registry image.
        image_name_src:  The source registry image.
        registry_v2: The underlying registry v2 image source to use to replicate.
    """
    docker_registry_client_async = registry_v2.docker_registry_client_async

    LOGGER.debug("Copying image: %s ...", image_name_src)

    manifest = await registry_v2.get_manifest(image_name=image_name_src)
    if isinstance(manifest, RegistryV2ManifestList):
        LOGGER.debug("Processing manifest list: %s ...", image_name_src)
        for digest in manifest.get_manifests():
            img_name_dest = image_name_dest.clone()
            img_name_dest.digest = digest
            img_name_dest.tag = None
            img_name_src = image_name_src.clone()
            img_name_src.digest = digest
            img_name_src.tag = None
            await copy_image(
                image_name_dest=img_name_dest,
                image_name_src=img_name_src,
                registry_v2=registry_v2,
            )
    else:
        for layer in manifest.get_layers():
            await copy_blob(
                digest=layer,
                image_name_dest=image_name_dest,
                image_name_src=image_name_src,
                registry_v2=registry_v2,
            )
        await copy_blob(
            digest=manifest.get_config_digest(),
            image_name_dest=image_name_dest,
            image_name_src=image_name_src,
            registry_v2=registry_v2,
        )

    await docker_registry_client_async.put_manifest(
        image_name=image_name_dest, manifest=manifest
    )


async def copy_manifest(
    *,
    image_name_dest: ImageName,
    image_name_src: ImageName,
    registry_v2: RegistryV2,
):
    """
    Replicates a manifest between two docker registry images.

    Args:
        image_name_dest: The destination registry image.
        image_name_src:  The source registry image.
        registry_v2: The underlying registry v2 image source to use to replicate.
    """
    docker_registry_client_async = registry_v2.docker_registry_client_async

    LOGGER.debug("Copying manifest: %s ...", image_name_src)
    response = await docker_registry_client_async.head_manifest(
        image_name=image_name_dest
    )
    if response.result:
        return

    async with aiotempfile(mode="w+b") as file:
        await docker_registry_client_async.get_manifest_to_disk(
            file=file, image_name=image_name_src
        )
        await be_kind_rewind(file)

        # Note: ClientResponse.content.iter_chunks() will deplete the underlying stream without saving
        #       ClientResponse._body; so calls to ClientReponse.read() will return None.
        # manifest = Manifest(await response["client_response"].read())
        manifest = Manifest(manifest=await file.read())
        await be_kind_rewind(file)

        response = await docker_registry_client_async.put_manifest_from_disk(
            file=file, image_name=image_name_dest, media_type=manifest.get_media_type()
        )
        # TODO: Is this check actually redundant to put_manifest_from_disk(check_digest=True)?
        must_be_equal(
            actual=response.digest,
            error_type=DigestMismatchError,
            expected=image_name_src.digest,
            msg="Manifest digest mismatch",
        )


def import_owner_trust(*, gpg: GPG, trust_data: str):
    # pylint: disable=protected-access
    """
    Imports trust information.

    Args:
        gpg: GPG object on which to operate.
        trust_data: The trust data to be imported.
    """
    result = gpg._result_map["import"](gpg)
    data = _make_binary_stream(trust_data, gpg._encoding)
    gpg._handle_io(["--import-ownertrust"], data, result, binary=True)
    data.close()
    return result


async def read_from_tar(tar_file, tarinfo: tarfile.TarInfo) -> bytes:
    """Reads an entry from a tar file into memory."""
    bytesio = BytesIO()
    await chunk_file(
        tar_file.extractfile(tarinfo),
        bytesio,
        file_in_is_async=False,
        file_out_is_async=False,
    )
    return bytesio.getvalue()


async def retrieve_and_verify_release_metadata(
    *,
    digest: FormattedSHA256,
    image_name: ImageName,
    keys: List[str],
    locations: List[str],
    registry_v2: RegistryV2,
) -> List[AtomicSignerVerify]:
    """
    Retrieves and verifies that a matching signatures exists for given digest / key combination at a set of predefined
    locations.

    Args:
        digest: The digest for which to verify the signature(s).
        image_name: Name of the image corresponding to the digest.
        keys: The public GPG keys to use to verify the signature.
        locations: The signature store locations at which to check for matching signature(s).
        registry_v2: The underlying registry v2 image source to use to verify the metadata.

    Returns:
        dict:
            result: Boolean result. True IFF a matching signature was found.
            signatures: List of matching signatures.
    """
    LOGGER.debug(
        "Verifying release authenticity:\nKeys      :\n  %s\nLocations :\n  %s",
        f"{len(keys)} key(s)",
        "\n  ".join(locations),
    )

    # TODO: Replace use of gnupg library with raw subprocess (and clean up associated log filtering) ...
    with TemporaryDirectory() as homedir:
        LOGGER.debug("Using trust store: %s", homedir)
        gpg = GPG(
            homedir=homedir,
            ignore_homedir_permissions=True,
            options=["--pinentry-mode loopback"],
        )

        LOGGER.debug("Importing keys ...")
        for key in keys:
            gpg.import_keys(key)

        # Would be nice if this method was built-in ... =/
        for key in gpg.list_keys():
            # TODO: Is it worth it to define pretty_bad_protocol._parsers.ImportOwnerTrust to validate
            #       the return value?
            import_owner_trust(gpg=gpg, trust_data=f"{key['fingerprint']}:6:\n")

        atomic_signer = AtomicSigner(
            docker_registry_client_async=registry_v2.docker_registry_client_async,
            homedir=homedir,
            locations=locations,
        )

        LOGGER.debug("Verifying signature(s) against digest: %s", digest)
        result = await atomic_signer.atomicverify(digest=digest, image_name=image_name)
        for signature in result:
            if not signature.valid:
                raise SignatureMismatchError("Signature does not match!")

            LOGGER.debug("Signature matches:")
            LOGGER.debug("  fingerprint : %s", signature.fingerprint)
            LOGGER.debug("  timestamp   : %s", signature.timestamp)
            LOGGER.debug("  username    : %s", signature.username)
            LOGGER.debug("Signature is compliant:")
            LOGGER.debug(
                "  type                   : %s", signature.atomic_signature.get_type()
            )
            LOGGER.debug(
                "  docker reference       : %s",
                signature.atomic_signature.get_docker_reference(),
            )
            LOGGER.debug(
                "  docker manifest digest : %s",
                signature.atomic_signature.get_docker_manifest_digest(),
            )
    if not result:
        raise NoSignatureError("Unable to locate a valid signature!")
    return result



#!/usr/bin/env python

# pylint: disable=too-many-arguments

"""docker-sign-verify command line interface."""

import logging
import os
import sys

from pathlib import Path
from traceback import print_exception
from typing import List, NamedTuple

import click

from click.core import Context
from docker_registry_client_async import ImageName
from docker_sign_verify import NoSignatureError, RegistryV2, SignatureMismatchError
from docker_sign_verify.scripts.utils import (
    async_command,
    HiddenPassword,
    LOGGING_DEFAULT,
    logging_options,
    set_log_levels,
    to_image_name,
)

from oc_mirror.atomicsigner import AtomicSignerVerify, AtomicSigner, FindAllSignatures

from .utils import OPENSHIFT_SIGNATURE_STORES, __version__

LOGGER = logging.getLogger(__name__)


class TypingContextObject(NamedTuple):
    # pylint: disable=missing-class-docstring
    registry_v2: RegistryV2
    signature_stores: List[str]
    signature_type: str
    signing_keys: List[str]
    verbosity: int


def get_context_object(*, context: Context) -> TypingContextObject:
    """Wrapper method to enforce type checking."""
    return context.obj


@click.group()
@click.option(
    "-s",
    "--signature-store",
    envvar="ATOMIC_SIGNATURE_STORE",
    help="Url of a signature store to use for retrieving signatures. Can be passed multiple times.",
    multiple=True,
)
@click.option(
    "-s",
    "--signature-type",
    envvar="ATOMIC_SIGNATURE_TYPE",
    help="Whether to sign the image configuration, or the registry manifest (atomic compatible).",
    default="image-config",
    type=click.Choice(["image-config", "manifest"], case_sensitive=False),
)
@click.option(
    "-k",
    "--signing-key",
    envvar="ATOMIC_SIGNING_KEY",
    help="Armored GnuPG trust store to use for signature verification. Can be passed multiple times.",
    multiple=True,
    type=click.Path(exists=True, dir_okay=False),
)
@logging_options
@click.pass_context
def cli(
    context: Context,
    signature_store: List[str],
    signature_type: str,
    signing_key: List[str],
    verbosity: int = LOGGING_DEFAULT,
):
    """Utility for creating and verifying atomic signatures."""

    if verbosity is None:
        verbosity = LOGGING_DEFAULT

    set_log_levels(verbosity)

    if not signature_store:
        signature_store = OPENSHIFT_SIGNATURE_STORES

    signing_keys = []
    for path in [Path(x) for x in signing_key]:
        LOGGER.debug("Loading signing key: %s", path)
        signing_keys.append(path.read_text("utf-8"))

    context.obj = TypingContextObject(
        registry_v2=RegistryV2(),
        signature_stores=signature_store,
        signature_type=signature_type,
        signing_keys=signing_keys,
        verbosity=verbosity,
    )


@cli.command()
@click.argument("image_name", callback=to_image_name, required=True)
@click.option(
    "-k",
    "--keyid",
    help="Signing key identifier.",
    required=True,
    envvar="ATOMIC_KEYID",
)
@click.option(
    "-p",
    "--keypass",
    default=lambda: HiddenPassword(os.environ.get("ATOMIC_KEYPASS", "")),
    help="Signing key passphrase.",
    hide_input=True,
    prompt=True,
)
@click.pass_context
@async_command
async def sign(
    context: Context,
    image_name: ImageName,
    keyid: str,
    keypass: str,
) -> FindAllSignatures:
    """Creates an atomic signature."""

    result = None
    ctx = get_context_object(context=context)
    try:
        atomic_signer = AtomicSigner(
            docker_registry_client_async=ctx.registry_v2.docker_registry_client_async,
            keyid=keyid,
            locations=ctx.signature_stores,
            passphrase=keypass,
        )
        digest = image_name.digest
        if not digest:
            LOGGER.debug(
                "Resolving digest for signature type: %s ...", ctx.signature_type
            )
            manifest = await ctx.registry_v2.get_manifest(image_name=image_name)
            digest = manifest.get_digest()
            LOGGER.debug("  manifest digest : %s", digest)
            if ctx.signature_type == "image-config":
                image_config = await ctx.registry_v2.get_image_config(
                    image_name=image_name
                )
                digest = image_config.get_digest()
                LOGGER.debug("  config digest   : %s", digest)
        result = await atomic_signer.atomicsign(digest=digest, image_name=image_name)
        if not result:
            raise RuntimeError(f"Unable to create new signature for digest: {digest}")
        LOGGER.info("Created new signature: %s (%s)", result.url, digest)
    except Exception as exception:  # pylint: disable=broad-except
        if ctx.verbosity > 0:
            logging.fatal(exception)
        if ctx.verbosity > LOGGING_DEFAULT:
            exc_info = sys.exc_info()
            print_exception(*exc_info)
        sys.exit(1)
    finally:
        await ctx.registry_v2.close()

    return result


@cli.command()
@click.argument("image_name", callback=to_image_name, nargs=-1, required=True)
@click.pass_context
@async_command
async def verify(
    context: Context,
    image_name: List[ImageName],
) -> List[List[AtomicSignerVerify]]:
    """Verifies an atomic signature(s)."""

    results = []
    ctx = get_context_object(context=context)
    try:
        atomic_signer = AtomicSigner(
            docker_registry_client_async=ctx.registry_v2.docker_registry_client_async,
            locations=ctx.signature_stores,
        )
        for img_name in image_name:
            LOGGER.info(img_name)
            digest = img_name.digest
            if not digest:
                LOGGER.debug(
                    "Resolving digest for signature type: %s ...", ctx.signature_type
                )
                manifest = await ctx.registry_v2.get_manifest(image_name=img_name)
                digest = manifest.get_digest()
                LOGGER.debug("  manifest digest : %s", digest)
                if ctx.signature_type == "image-config":
                    image_config = await ctx.registry_v2.get_image_config(
                        image_name=img_name
                    )
                    digest = image_config.get_digest()
                    LOGGER.debug("  config digest   : %s", digest)
            result = await atomic_signer.atomicverify(
                digest=digest, image_name=img_name
            )
            for signature in result:
                if not signature.valid:
                    raise SignatureMismatchError("Signature does not match!")

                LOGGER.debug("Signature matches:")
                LOGGER.debug("  fingerprint : %s", signature.fingerprint)
                LOGGER.debug("  timestamp   : %s", signature.timestamp)
                LOGGER.debug("  username    : %s", signature.username)
                LOGGER.debug("Signature is compliant:")
                LOGGER.debug(
                    "  type                   : %s",
                    signature.atomic_signature.get_type(),
                )
                LOGGER.debug(
                    "  docker reference       : %s",
                    signature.atomic_signature.get_docker_reference(),
                )
                LOGGER.debug(
                    "  docker manifest digest : %s",
                    signature.atomic_signature.get_docker_manifest_digest(),
                )
            if not result:
                raise NoSignatureError("Unable to locate a valid signature!")
            LOGGER.info("Verified %s signatures.", len(result))
            results.append(result)
    except Exception as exception:  # pylint: disable=broad-except
        if ctx.verbosity > 0:
            logging.fatal(exception)
        if ctx.verbosity > LOGGING_DEFAULT:
            exc_info = sys.exc_info()
            print_exception(*exc_info)
        sys.exit(1)
    finally:
        await ctx.registry_v2.docker_registry_client_async.close()

    return results


@click.command()
def version():
    """Displays the utility version."""
    print(__version__)


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    cli()



#!/usr/bin/env python

# pylint: disable=redefined-outer-name

"""OpenShift release helpers."""

import gzip
import logging
import re
import tarfile

from json import loads
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Set, Tuple
from yaml import load_all, SafeLoader

from aiotempfile.aiotempfile import open as aiotempfile
from docker_registry_client_async import (
    FormattedSHA256,
    ImageName,
)
from docker_sign_verify import ImageConfig, RegistryV2, RegistryV2ManifestList
from docker_sign_verify.utils import be_kind_rewind

from .atomicsigner import AtomicSignerVerify
from .imagestream import ImageStream
from .singleassignment import SingleAssignment
from .specs import OpenShiftReleaseSpecs
from .utils import (
    copy_blob,
    copy_manifest,
    read_from_tar,
    TypingRegexSubstitution,
    retrieve_and_verify_release_metadata,
)

LOGGER = logging.getLogger(__name__)


class TypingCollectDigests(NamedTuple):
    # pylint: disable=missing-class-docstring
    blobs: Dict[FormattedSHA256, Set[str]]
    manifests: Dict[ImageName, str]


class TypingGetReleaseMetadata(NamedTuple):
    # pylint: disable=missing-class-docstring
    blobs: Dict[FormattedSHA256, Set[str]]
    manifest_digest: FormattedSHA256
    manifests: Dict[ImageName, str]
    raw_image_references: ImageStream
    raw_release_metadata: Any
    signature_stores: List[str]
    signatures: List[AtomicSignerVerify]
    signing_keys: List[str]


class TypingGetSecurityInformation(NamedTuple):
    # pylint: disable=missing-class-docstring
    keys: List[str]
    locations: List[str]


class TypingSearchLayer(NamedTuple):
    # pylint: disable=missing-class-docstring
    image_references: ImageStream
    keys: List[str]
    locations: List[str]
    release_metadata: Any


async def _collect_digests(
    *,
    image_references: ImageStream,
    registry_v2: RegistryV2,
    release_name: ImageName,
) -> TypingCollectDigests:
    """
    Retrieves all blob and manifest digests for a given release.

    Args:
        image_references: The image references for the release.
        registry_v2: The underlying registry v2 image source to use to retrieve the digests.
        release_name: The name of the release image.

    Returns:
        blobs: The mapping of blob digests to image prefixes.
        manifests: The mapping of image manifests to image stream names.
    """
    blobs = {}
    manifests = {}

    def add_blob(_digest: FormattedSHA256, i_prefix: str):
        if _digest not in blobs:
            blobs[_digest] = set()
        blobs[_digest].add(i_prefix)

    # TODO: Should we split out manifest and blob processing to separate functions?
    for image_name, name in _get_tag_mapping(
        release_name=release_name, image_references=image_references
    ):
        # Convert tags to digests
        # pkg/cli/image/mirror/mirror.go:437 - plan()
        digest = image_name.digest
        if image_name.tag and not image_name.digest:
            response = await registry_v2.docker_registry_client_async.head_manifest(
                image_name
            )
            LOGGER.debug("Resolved source image %s to %s", image_name, response.digest)
            digest = response.digest

        # Find all blobs ...
        # TODO: Handle manifest lists ...
        manifest = await registry_v2.get_manifest(image_name=image_name)
        image_prefix = f"{image_name.endpoint}/{image_name.image}"
        add_blob(manifest.get_config_digest(), image_prefix)
        for layer in manifest.get_layers():
            add_blob(layer, image_prefix)

        # Note: Must be assigned below blob inspection to prevent errors based on digest lookup
        image_name.digest = digest
        image_name.tag = ""
        manifests[image_name] = name

    return TypingCollectDigests(blobs=blobs, manifests=manifests)


async def _get_image_references(
    *, tar_file, tarinfo: tarfile.TarInfo, path: Path
) -> Optional[ImageStream]:
    """
    Retrieves images references from a given tarinfo, if available.

    Args:
        tar_file: The tar_file of the layer being processed.
        tarinfo: The tarinfo from which to retrieve the image references.
        path: Relative path of the tarinfo within the tar file.

    Returns:
        An ImageStream object containing the image references, or None.
    """
    # pkg/cli/admin/release/mirror.go:475 - Run()
    if path.name == OpenShiftReleaseSpecs.IMAGE_REFERENCES_NAME:
        LOGGER.debug("Found image references: %s", path)
        result = await read_from_tar(tar_file, tarinfo)
        return ImageStream(result)


async def _get_release_metadata(
    *, tar_file, tarinfo: tarfile.TarInfo, path: Path
) -> Optional[Any]:
    """
    Retrieves release metadata from a given tarinfo, if available.

    Args:
        tar_file: The tar_file of the layer being processed.
        tarinfo: The tarinfo from which to retrieve the release metadata.
        path: Relative path of the tarinfo within the tar file.

    Returns:
        The release metadata data structure, or None.
    """
    # pkg/cli/admin/release/mirror.go:475 - Run()
    if path.name == OpenShiftReleaseSpecs.RELEASE_METADATA:
        LOGGER.debug("Found release metadata: %s", path)
        result = await read_from_tar(tar_file, tarinfo)
        return loads(result)


async def _get_security_information(
    *, tar_file, tarinfo: tarfile.TarInfo, path: Path
) -> TypingGetSecurityInformation:
    """
    Retrieves security information from a given tarinfo, if available.

    Args:
        tar_file: The tar_file of the layer being processed.
        tarinfo: The tarinfo from which to retrieve the security information.
        path: Relative path of the tarinfo within the tar file.

    Returns:
        keys: The list of public GPG keys found within the tarinfo.
        locations: The list of signature store locations found within the tarinfo.
    """
    keys = []
    locations = []

    # pkg/cli/admin/release/extract.go:228 - Run()
    if path.suffix in [".yaml", ".yml", ".json"]:
        # LOGGER.debug("Found manifest: %s", path_file.name)

        # ... for all matching files found, parse them ...
        _bytes = await read_from_tar(tar_file, tarinfo)
        if path.suffix == ".json":
            documents = [loads(_bytes)]
        else:
            documents = load_all(_bytes, SafeLoader)

        # ... and look for a root-level "data" key ...
        for document in documents:
            if not document:
                continue
            if document.get("kind", "") != "ConfigMap":
                continue
            if (
                OpenShiftReleaseSpecs.RELEASE_ANNOTATION_CONFIG_MAP_VERIFIER
                not in document.get("metadata", []).get("annotations", [])
            ):
                continue
            LOGGER.debug("Found release security information: %s", path)
            for key, value in document.get("data", {}).items():
                if key.startswith("verifier-public-key-"):
                    # LOGGER.debug("Found in %s:\n%s %s", path.name, key, value)
                    keys.append(value)
                if key.startswith("store-"):
                    # LOGGER.debug("Found in %s:\n%s\n%s", path.name, key, value)
                    locations.append(value)
    return TypingGetSecurityInformation(keys=keys, locations=locations)


def _get_tag_mapping(
    *, release_name: ImageName, image_references: ImageStream
) -> Tuple[ImageName, str]:
    """
    Deconstructs the metadata inside an ImageStream into a mapping of image names to tag names.

    Args:
        release_name: The name of the release image.
        image_references: The image references for the release.
    Yields:
        A tuple of image name and tag name.
    """
    # Special Case: for the outer release image
    # pkg/cli/admin/release/mirror.go:565 (o.ToRelease mapping)
    yield release_name.clone(), release_name.tag

    for name, image_name in image_references.get_tags():
        assert not image_name.tag and image_name.digest
        # LOGGER.debug("Mapping %s -> %s", image_name, name)
        yield image_name, name


async def _search_layer(
    *,
    layer: FormattedSHA256,
    registry_v2: RegistryV2,
    release_name: ImageName,
) -> TypingSearchLayer:
    """
    Searches image layers in a given release image for metadata.

    Args:
        layer: The image layer to be searched.
        registry_v2: The underlying registry v2 image source to use to retrieve the metadata.
        release_name: The name of the release image.

    Returns:
        image_references: An ImageStream object containing the image references, or None.
        keys: The list of public GPG keys found within the layer.
        locations: The list of signature store locations found within the layer.
        release_metadata: The release metadata data structure, or None.
    """
    LOGGER.debug("Extracting from layer : %s", layer)

    image_references = SingleAssignment("image_references")
    keys = []
    locations = []
    release_metadata = SingleAssignment("release_metadata")
    async with aiotempfile(mode="w+b") as file:
        await registry_v2.docker_registry_client_async.get_blob_to_disk(
            digest=layer, file=file, image_name=release_name
        )
        await be_kind_rewind(file)

        with gzip.GzipFile(filename=file.name) as gzip_file_in:
            with tarfile.open(fileobj=gzip_file_in) as tar_file_in:
                for tarinfo in tar_file_in:
                    path = Path(tarinfo.name)

                    # TODO: Do we need to process tarinfo.linkname like pkg/cli/image/extract/extract.go:621 ?

                    if not str(path).startswith(
                        OpenShiftReleaseSpecs.MANIFEST_PATH_PREFIX
                    ):
                        # pkg/cli/image/extract/extract.go:616 - changeTarEntryParent()
                        # LOGGER.debug("Exclude %s due to missing prefix %s", path)
                        continue
                    tmp = await _get_image_references(
                        tar_file=tar_file_in, tarinfo=tarinfo, path=path
                    )
                    if tmp:
                        image_references.set(tmp)
                    tmp = await _get_release_metadata(
                        tar_file=tar_file_in, tarinfo=tarinfo, path=path
                    )
                    if tmp:
                        release_metadata.set(tmp)
                    tmp = await _get_security_information(
                        tar_file=tar_file_in, tarinfo=tarinfo, path=path
                    )
                    keys.extend(tmp.keys)
                    locations.extend(tmp.locations)
    return TypingSearchLayer(
        image_references=image_references.get(),
        keys=keys,
        locations=locations,
        release_metadata=release_metadata.get(),
    )


async def get_release_metadata(
    *,
    registry_v2: RegistryV2,
    release_name: ImageName,
    signature_stores: List[str] = None,
    signing_keys: List[str] = None,
    verify: bool = True,
) -> TypingGetReleaseMetadata:
    # pylint: disable=protected-access,too-many-locals
    """
    Retrieves all metadata for a given OpenShift release image.

    Args:
        registry_v2: The Registry V2 image source to use to connect.
        release_name: The OpenShift release image for which to retrieve the metadata.
        signature_stores: A list of signature store uri overrides.
        signing_keys: A list of armored GnuPG trust store overrides.
        verify: If True, the atomic signature will be retrieved and validated.

    Returns:
        dict:
            blobs: A mapping of blob digests to a set of image prefixes.
            manifests: A mapping of image manifests to tag values.
            signature_stores: A list of signature store uris.
            signing_keys: A list of armored GnuPG trust stores.
    """
    # TODO: Change assertions to runtime checks.
    LOGGER.debug("Source release image name: %s", release_name)

    # pkg/cli/image/extract/extract.go:332 - Run()
    # pkg/cli/image/manifest/manifest.go:342 - ProcessManifestList()

    # Retrieve the manifest ...
    manifest = await registry_v2.get_manifest(image_name=release_name)
    if isinstance(manifest, RegistryV2ManifestList):
        LOGGER.debug("Manifest list digest: %s", manifest.get_digest())
        manifest = await registry_v2._get_manifest_from_manifest_list(
            image_name=release_name, manifest_list=manifest
        )
    manifest_digest = manifest.get_digest()
    LOGGER.debug("Source release manifest digest: %s", manifest_digest)

    # TODO: Do we need pkg/cli/image/manifest/manifest.go:70 - Verify() ?

    # Log the image configuration (but why?)
    response = await registry_v2.docker_registry_client_async.get_blob(
        release_name, manifest.get_config_digest()
    )
    assert response.blob
    image_config = ImageConfig(response.blob)
    # pkg/cli/image/manifest/manifest.go:289 - ManifestToImageConfig()
    LOGGER.debug("Source release image config digest: %s", image_config.get_digest())

    # Search through all layers in reverse order, looking for yaml and json files under a given prefix ...
    # pkg/cli/image/extract/extract.go:307 - Run()
    image_references = SingleAssignment("image_references")
    keys = []
    locations = []
    release_metadata = SingleAssignment("release_metadata")
    for layer in manifest.get_layers():
        tmp = await _search_layer(
            layer=FormattedSHA256.parse(layer),
            registry_v2=registry_v2,
            release_name=release_name,
        )
        if tmp.image_references:
            image_references.set(tmp.image_references)
        keys.extend(tmp.keys)
        locations.extend(tmp.locations)
        if tmp.release_metadata:
            release_metadata.set(tmp.release_metadata)
    image_references = image_references.get()
    release_metadata = release_metadata.get()

    assert image_references
    assert keys
    assert locations
    assert release_metadata

    # pkg/cli/admin/release/mirror.go:517 - imageVerifier.Verify()
    signatures = []
    if verify:
        _keys = signing_keys if signing_keys else keys
        _locations = signature_stores if signature_stores else locations
        signatures = await retrieve_and_verify_release_metadata(
            digest=manifest_digest,
            image_name=release_name,
            keys=_keys,
            locations=_locations,
            registry_v2=registry_v2,
        )
    else:
        LOGGER.debug("Skipping source release authenticity verification!")

    assert image_references.get_json().get("kind", "") == "ImageStream"
    assert image_references.get_json().get("apiVersion", "") == "image.openshift.io/v1"

    tmp = await _collect_digests(
        image_references=image_references,
        registry_v2=registry_v2,
        release_name=release_name,
    )
    LOGGER.debug(
        "Collected %d manifests with %d blobs.",
        len(tmp.manifests),
        len(tmp.blobs),
    )
    result = TypingGetReleaseMetadata(
        blobs=tmp.blobs,
        manifest_digest=manifest_digest,
        manifests=tmp.manifests,
        raw_image_references=image_references,
        raw_release_metadata=release_metadata,
        signature_stores=locations,
        signatures=signatures,
        signing_keys=keys,
    )

    # pkg/cli/image/mirror/plan.go:244 - Print()
    # LOGGER.debug(release_name)
    # LOGGER.debug("  blobs:")
    # for digest, image_prefixes in result.blobs.items():
    #     for image_prefix in image_prefixes:
    #         LOGGER.debug("    %s %s", image_prefix, digest)
    # LOGGER.debug("  manifests:")
    # for image_name in result.manifests.keys():
    #     LOGGER.debug("    %s -> %s", image_name, result.manifests[image_name])

    return result


async def log_release_metadata(
    *,
    release_name: ImageName,
    release_metadata: TypingGetReleaseMetadata,
    sort_metadata: bool = False,
):
    """
    Appends metadata for a given release to the log

    Args:
        release_name: The OpenShift release image for which to retrieve the metadata.
        release_metadata: The metadata for the release to be logged.
        sort_metadata: If True, the metadata keys will be sorted.
    """

    def make_image_name(img_name: ImageName, tag: str):
        result = img_name.clone()
        result.tag = tag
        return result

    LOGGER.info(release_name)
    LOGGER.info("  manifests:")
    manifests = [
        make_image_name(image_name, tag)
        for image_name, tag in release_metadata.manifests.items()
    ]
    if sort_metadata:
        manifests = sorted(manifests)
    for manifest in manifests:
        LOGGER.info("    %s", manifest)
    LOGGER.info("  blobs:")
    blobs = [
        ImageName.parse(f"{image_prefix}@{digest}")
        for digest, image_prefixes in release_metadata.blobs.items()
        for image_prefix in image_prefixes
    ]
    if sort_metadata:
        blobs = sorted(blobs)
    for blob in blobs:
        LOGGER.info("    %s", blob)
    LOGGER.info("  signatures:")
    signatures = release_metadata.signatures
    if sort_metadata:
        signatures = sorted(signatures, key=lambda item: item.timestamp, reverse=True)
    for signature in signatures:
        LOGGER.info(
            "    release name    : %s",
            signature.atomic_signature.get_docker_reference(),
        )
        LOGGER.info(
            "    manifest digest : %s",
            signature.atomic_signature.get_docker_manifest_digest(),
        )
        LOGGER.info("    fingerprint     : %s", signature.key_id)
        LOGGER.info("    username        : %s", signature.username)
        LOGGER.info("    timestamp       : %s", signature.timestamp)
        LOGGER.info("")


async def put_release(
    *,
    registry_v2: RegistryV2,
    release_metadata: TypingGetReleaseMetadata,
    release_name: ImageName,
    verify: bool = True,
):
    """
    Mirrors an openshift release.

    Args:
        registry_v2: The Registry V2 image source to use to connect.
        release_metadata: The metadata for the release to be mirrored.
        release_name: The OpenShift release image to which to store the metadata.
        verify: If True, the atomic signature will be retrieved and validated.
    """
    LOGGER.debug("Destination release image name: %s", release_name)

    LOGGER.debug(
        "Replicating %d manifests with %d blobs.",
        len(release_metadata.manifests),
        len(release_metadata.blobs),
    )

    if verify:
        await retrieve_and_verify_release_metadata(
            digest=release_metadata.manifest_digest,
            image_name=release_name,
            keys=release_metadata.signing_keys,
            locations=release_metadata.signature_stores,
            registry_v2=registry_v2,
        )
    else:
        LOGGER.debug("Skipping source release authenticity verification!")

    # Replicate all blobs ...
    last_image_name_dest = None
    last_image_name_src = None
    for digest, image_prefixes in release_metadata.blobs.items():
        # TODO: Handle blob mounting ...
        for image_prefix in image_prefixes:
            image_name_src = ImageName.parse(image_prefix)
            image_name_dest = image_name_src.clone()
            # Note: Only update the endpoint; keep the digest and image the same
            image_name_dest.endpoint = release_name.endpoint

            if (
                last_image_name_dest != image_name_dest
                or last_image_name_src != image_name_src
            ):
                LOGGER.debug("Copy blobs ...")
                LOGGER.debug("    from : %s", image_name_src)
                LOGGER.debug("    to   : %s", image_name_dest)
            await copy_blob(
                digest=digest,
                image_name_dest=image_name_dest,
                image_name_src=image_name_src,
                registry_v2=registry_v2,
            )
            last_image_name_dest = image_name_dest
            last_image_name_src = image_name_src

    # Replicate all manifests ...
    for image_name_src in release_metadata.manifests.keys():
        # Note: Update the endpoint; keep the image unchanged; use the derived tag; do not use digest
        image_name_dest = ImageName(
            image_name_src.image,
            endpoint=release_name.endpoint,
            tag=release_metadata.manifests[image_name_src],
        )
        await copy_manifest(
            image_name_dest=image_name_dest,
            image_name_src=image_name_src,
            registry_v2=registry_v2,
        )


async def translate_release_metadata(
    *,
    regex_substitutions: List[TypingRegexSubstitution],
    release_metadata: TypingGetReleaseMetadata,
    signature_stores: List[str] = None,
    signing_keys: List[str] = None,
) -> TypingGetReleaseMetadata:
    """
    Translates metadata for a given release to support hop 2-n mirroring.

    Args:
        regex_substitutions: Regular expression substitutions to be applied to source uris.
        release_metadata: The metadata for the release to be translated.
        signature_stores: A list of signature store uri overrides.
        signing_keys: A list of armored GnuPG trust store overrides.

    Returns:
        The translated metadata for a given release.
    """
    if signature_stores is None:
        signature_stores = release_metadata.signature_stores
    if signing_keys is None:
        signing_keys = release_metadata.signing_keys

    blobs = release_metadata.blobs
    manifests = release_metadata.manifests
    for regex_substitution in regex_substitutions:
        pattern = re.compile(regex_substitution.pattern)
        blobs = {
            k: {pattern.sub(regex_substitution.replacement, x) for x in v}
            for (k, v) in release_metadata.blobs.items()
        }
        manifests = {
            ImageName.parse(pattern.sub(regex_substitution.replacement, str(k))): v
            for (k, v) in release_metadata.manifests.items()
        }

    return TypingGetReleaseMetadata(
        blobs=blobs,
        manifest_digest=release_metadata.manifest_digest,
        manifests=manifests,
        raw_image_references=release_metadata.raw_image_references,
        raw_release_metadata=release_metadata.raw_release_metadata,
        signatures=release_metadata.signatures,
        signature_stores=signature_stores,
        signing_keys=signing_keys,
    )



#!/usr/bin/env python

# pylint: disable=too-few-public-methods

"""Reusable string literals."""


class OpenShiftReleaseSpecs:
    """
    OpenShift string literals.
    """

    IMAGE_REFERENCES_NAME = "image-references"
    MANIFEST_PATH_PREFIX = "release-manifests/"
    RELEASE_ANNOTATION_CONFIG_MAP_VERIFIER = (
        "release.openshift.io/verification-config-map"
    )
    RELEASE_METADATA = "release-metadata"


class OperatorReleaseSpecs:
    """
    Operator string literals.
    """

    DATABASE_PATH_PREFIX = "database/"
    INDEX_DATABASE_NAME = "index.db"



#!/usr/bin/env python

"""Utility classes."""

import click

from oc_mirror import __version__

OPENSHIFT_SIGNATURE_STORES = [
    "https://mirror.openshift.com/pub/openshift-v4/signatures/openshift/release"
]


@click.command()
def version():
    """Displays the utility version."""
    print(__version__)



#!/usr/bin/env python

"""Manifest tests."""

import logging

from _pytest.logging import LogCaptureFixture

from oc_mirror.singleassignment import SingleAssignment


def test_get_set(caplog: LogCaptureFixture):
    """Tests double assignment."""

    caplog.clear()
    caplog.set_level(logging.DEBUG)

    name = "myvarname"
    value1 = "first value"
    single_assignment = SingleAssignment(name)
    single_assignment.set(value1)
    assert single_assignment.get() == value1
    assert "Overriding previously value of" not in caplog.text

    value2 = "second value"
    single_assignment.set(value2)
    assert "Overriding previously value of" in caplog.text
    assert name in caplog.text

    caplog.clear()

    assert single_assignment.get() == value2
    assert "Overriding previously value of" not in caplog.text


def test_accessor_methods(caplog: LogCaptureFixture):
    """Tests double assignment."""

    caplog.clear()
    caplog.set_level(logging.DEBUG)

    name = "myvarname"
    value1 = "first value"
    single_assignment = SingleAssignment(name)
    single_assignment.myvarname = value1
    assert single_assignment.myvarname == value1
    assert "Overriding previously value of" not in caplog.text

    value2 = "second value"
    single_assignment.myvarname = value2
    assert "Overriding previously value of" in caplog.text
    assert name in caplog.text

    caplog.clear()

    assert single_assignment.myvarname == value2
    assert "Overriding previously value of" not in caplog.text



#!/usr/bin/env python

"""Utility classes."""

import logging
import shlex
import sys

from functools import wraps
from pathlib import Path
from typing import Union

import pytest

from click.testing import CliRunner, Result
from docker_registry_client_async import ImageName

LOGGER = logging.getLogger(__name__)


class OCMCliRunner(CliRunner):
    """
    click.testing.CliRunner.invoke w/o isolation, as it breaks logging:
    https://github.com/pallets/click/issues/824
    """

    def invoke(self, cli, args=None, catch_exceptions=True, **extra) -> Result:
        # pylint: disable=arguments-differ
        exc_info = None
        exception = None
        exit_code = 0

        if isinstance(args, str):
            args = shlex.split(args)

        if "prog_name" not in extra:
            extra["prog_name"] = self.get_default_prog_name(cli)

        try:
            cli.main(args=args or (), **extra)
        except SystemExit as exc:
            exc_info = sys.exc_info()
            exit_code = exc.code
            if exit_code is None:
                exit_code = 0

            if exit_code != 0:
                exception = exc

            if not isinstance(exit_code, int):
                sys.stdout.write(str(exit_code))
                sys.stdout.write("\n")
                exit_code = 1

        except Exception as exc:  # pylint: disable=broad-except
            if not catch_exceptions:
                raise
            exc_info = sys.exc_info()
            exit_code = 1
            exception = exc

        return Result(
            exception=exception,
            exit_code=exit_code,
            exc_info=exc_info,
            return_value=exit_code,
            runner=self,
            stderr_bytes=b"",
            stdout_bytes=b"",
        )


def equal_if_unqualified(image_name0: ImageName, image_name1: ImageName) -> bool:
    """
    Determines if two images names are equal if evaluated as unqualified.

    Args:
        image_name0: The name of the first image.
        image_name1: The name of the second image.

    Returns:
        True if the images names are equal without considering the endpoint component.
    """
    img_name0 = image_name0.clone()
    img_name0.endpoint = None
    img_name1 = image_name1.clone()
    img_name1.endpoint = None
    return str(img_name0) == str(img_name1)


def get_test_data_path(request, name) -> Path:
    """Helper method to retrieve the path of test data."""
    return Path(request.fspath).parent.joinpath("data").joinpath(name)


def get_test_data(request, klass, name, mode="rb") -> Union[bytes, str]:
    # pylint: disable=unspecified-encoding
    """Helper method to retrieve test data."""
    key = f"{klass}/{name}"
    result = request.config.cache.get(key, None)
    if result is None:
        path = get_test_data_path(request, name)
        with open(path, mode) as file:
            result = file.read()
            # TODO: How do we / Should we serialize binary data?
            # request.config.cache.set(key, result)
    return result


def needs_credentials(*credentials):
    """Validates the state of RegistryV2.docker_registry_client_async.credentials before invoking the wrapped method."""

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            param = "registry_v2"
            if "registry_v2" in kwargs:
                ...
                for credential in credentials:
                    if credential not in str(
                        kwargs[param].docker_registry_client_async.credentials
                    ):
                        pytest.skip(
                            f"RegistryV2 does not contain required credentials for: {credential}! "
                            "Verify the default docker credentials store, or the location specified "
                            "by DRCA_CREDENTIALS_STORE."
                        )
            else:
                LOGGER.warning(
                    "Decorator %s used on %s without parameter %s!",
                    __name__,
                    func,
                    param,
                )
            return await func(*args, **kwargs)

        return wrapper

    return decorator



#!/usr/bin/env python

"""Classes that provide signature functionality."""

import logging

from pathlib import Path
from typing import List, NamedTuple, Optional
from urllib.parse import urlparse

import aiofiles

from aiohttp.client import _RequestContextManager
from aiohttp.typedefs import LooseHeaders
from aiotempfile.aiotempfile import open as aiotempfile
from docker_registry_client_async import (
    DockerRegistryClientAsync,
    FormattedSHA256,
    ImageName,
)
from docker_sign_verify import (
    DigestMismatchError,
    GPGSigner,
    GPGStatus,
    GPGTrust,
    SignatureMismatchError,
)

from .atomicsignature import AtomicSignature

LOGGER = logging.getLogger(__name__)


class AtomicSignerVerify(NamedTuple):
    # pylint: disable=missing-class-docstring
    atomic_signature: AtomicSignature
    fingerprint: str
    key_id: str
    signer_long: Optional[str]
    signer_short: Optional[str]
    status_atomic: Optional[ValueError]
    status_gpg: str
    timestamp: str
    trust: str
    type: str
    username: str
    valid: bool


class FindAllSignatures(NamedTuple):
    # pylint: disable=missing-class-docstring
    index: int
    signature: bytes
    url: str


async def get_request_headers(*, headers: LooseHeaders = None) -> LooseHeaders:
    """
    Generates request headers that contain the user agent identifier.

    Args:
        headers: Optional supplemental request headers to be returned.

    Returns:
        The generated request headers.
    """
    if not headers:
        headers = {}

    if "User-Agent" not in headers:
        # Note: This cannot be imported above, as it causes a circular import!
        from . import __version__  # pylint: disable=import-outside-toplevel

        headers["User-Agent"] = f"oc-mirror/{__version__}"

    return headers


class AtomicSigner(GPGSigner):
    """Creates and verifies atomic signatures using GnuPG."""

    def __init__(
        self,
        *,
        docker_registry_client_async: DockerRegistryClientAsync,
        homedir: Path = None,
        keyid: str = None,
        locations: List[str] = None,
        passphrase: str = None,
    ):
        """
        Args:
            docker_registry_client_async: The underlying DockerRegistryClientAsync object.
            homedir: The GPG home directory (default: ~/.gnupg).
            keyid: The GPG key identifier, only required for signing.
            passphrase: The passphrase used to unlock the GPG key.
            locations: The signature store locations used to retrieve signatures.
        """
        super().__init__(homedir=homedir, keyid=keyid, passphrase=passphrase)

        self.docker_registry_client_async = docker_registry_client_async
        self.locations = locations if locations is not None else []

        LOGGER.debug("Using signature stores:")
        for location in locations:
            LOGGER.debug("  %s", location)

    async def _find_all_signatures(
        self, *, digest: FormattedSHA256, location: str
    ) -> List[FindAllSignatures]:
        # pylint: disable=protected-access
        """Retrieves the all signatures corresponding to a given digest from a signature store."""
        result = []
        client_session = await self.docker_registry_client_async._get_client_session()
        headers = await self._get_headers(endpoint=location)
        proxy = await self._get_proxy(location=location)
        index = 0
        while True:
            index = index + 1
            url = f"{location}/sha256={digest.sha256}/signature-{index}"
            LOGGER.debug("Retrieving signature: %s ...", url)
            response = await client_session.get(
                headers=headers,
                proxy=proxy,
                proxy_auth=self.docker_registry_client_async.proxy_auth,
                ssl=self.docker_registry_client_async.ssl,
                url=url,
            )
            if response.status > 400:
                break
            LOGGER.debug("Signature retrieved.")
            result.append(
                FindAllSignatures(index=index, signature=await response.read(), url=url)
            )
        return result

    async def _get_headers(self, *, endpoint: str) -> LooseHeaders:
        # pylint: disable=protected-access
        """Retrieves headers for remote requests."""
        url_segments = urlparse(endpoint)
        for i in [
            url_segments.netloc,
            f"{url_segments.scheme}://{url_segments.netloc}",
            f"{url_segments.scheme}://{url_segments.netloc}/",
            f"{url_segments.netloc}:{url_segments.port}",
            f"{url_segments.scheme}://{url_segments.netloc}:{url_segments.port}",
            f"{url_segments.scheme}://{url_segments.netloc}:{url_segments.port}/",
        ]:
            credentials = await self.docker_registry_client_async._get_credentials(
                endpoint=i
            )
            if credentials:
                break
        headers = await get_request_headers()
        if credentials:
            headers["Authorization"] = f"Basic {credentials}"
        return headers

    async def _get_proxy(self, *, location: str) -> Optional[str]:
        # pylint: disable=protected-access
        """
        Retrieves the proxy configuration for a given location.

        Args:
            location: The location for which to retrieve the proxy configuration.
        """
        url_segments = urlparse(location)
        proxy = await self.docker_registry_client_async._get_proxy(
            endpoint=url_segments.netloc, protocol=url_segments.scheme.lower()
        )
        return proxy

    async def _make_collection(self, *, digest: FormattedSHA256, location: str):
        # pylint: disable=protected-access
        """Creates a remote collection (directory), if needed, for a given digest."""
        client_session = await self.docker_registry_client_async._get_client_session()
        headers = await self._get_headers(endpoint=location)
        proxy = await self._get_proxy(location=location)
        url = f"{location}/sha256={digest.sha256}"
        response = await client_session.get(
            headers=headers,
            proxy=proxy,
            proxy_auth=self.docker_registry_client_async.proxy_auth,
            ssl=self.docker_registry_client_async.ssl,
            url=url,
        )
        if response.status < 400:
            return
        LOGGER.debug("Creating collection: %s ...", url)
        await _RequestContextManager(
            client_session._request(
                allow_redirects=True,
                headers=headers,
                method="MKCOL",
                proxy=proxy,
                proxy_auth=self.docker_registry_client_async.proxy_auth,
                raise_for_status=True,
                ssl=self.docker_registry_client_async.ssl,
                str_or_url=url,
            )
        )
        LOGGER.debug("Collection created.")

    async def _store_signature(
        self, *, digest: FormattedSHA256, location: str = None, signature: bytes
    ) -> FindAllSignatures:
        # pylint: disable=protected-access
        """Replicates a given signature to a remote signature store."""
        if location is None:
            location = self.locations[0]

        signatures = await self._find_all_signatures(digest=digest, location=location)
        index = signatures[-1].index + 1 if signatures else 1
        url = f"{location}/sha256={digest.sha256}/signature-{index}"

        await self._make_collection(digest=digest, location=location)

        LOGGER.debug("Storing signature: %s ...", url)
        client_session = await self.docker_registry_client_async._get_client_session()
        headers = await self._get_headers(endpoint=location)
        proxy = await self._get_proxy(location=location)
        await client_session.put(
            data=signature,
            headers=headers,
            proxy=proxy,
            proxy_auth=self.docker_registry_client_async.proxy_auth,
            raise_for_status=True,
            ssl=self.docker_registry_client_async.ssl,
            url=url,
        )
        LOGGER.debug("Signature stored.")
        return FindAllSignatures(index=index, signature=signature, url=url)

    @staticmethod
    async def _verify_atomic_signature(
        *,
        atomic_signature: AtomicSignature,
        digest: FormattedSHA256,
        image_name: ImageName,
    ) -> Optional[ValueError]:
        """
        Verifies an atomic signature against a given digest and image name.
        pkg/verify/verify.go:382: - verifyAtomicContainerSignature()
        """
        if atomic_signature.get_type() != AtomicSignature.TYPE:
            return SignatureMismatchError("Signature is not the correct type!")
        # TODO: Actually make use of the provided image_name ...
        if len(atomic_signature.get_docker_reference().image) < 1:
            return SignatureMismatchError("Signature must have an identity!")
        if atomic_signature.get_docker_manifest_digest() != digest:
            return DigestMismatchError("Signature digest does not match!")
        return None

    async def atomicsign(
        self, *, digest: FormattedSHA256, image_name: ImageName
    ) -> Optional[FindAllSignatures]:
        """Creates an atomic signature."""
        atomic_signature = AtomicSignature.minimal(
            docker_manifest_digest=digest, docker_reference=str(image_name)
        )

        if not self.keyid:
            raise RuntimeError("Cannot sign without keyid!")
        if not self.passphrase or len(self.passphrase) < 1:
            raise RuntimeError("Refusing to use an unprotected key!")

        # Write the data to a temporary file and invoke GnuPG to create a detached signature ...
        async with aiotempfile(mode="w+b") as datafile:
            signaturefile = Path(f"{datafile.name}.gpg")

            # Write the data to a temporary file ...
            await datafile.write(atomic_signature.get_bytes())
            await datafile.flush()

            # Note: These options deviate from GPGSigner.sign()
            args = [
                "gpg",
                "--batch",
                "--default-key",
                str(self.keyid),
                "--digest-algo",
                "SHA512",
                "--homedir",
                str(self.homedir),
                "--no-options",
                "--no-emit-version",
                "--no-tty",
                "--passphrase-fd",
                "0",
                "--pinentry-mode",
                "loopback",
                "--sign",
                "--status-fd",
                "2",
                datafile.name,
            ]

            execute_command = await self._exeute_command(
                args=args, stdin=self.passphrase.encode("utf-8")
            )
            if execute_command.returncode:
                return None

        # Retrieve the signature and cleanup ...
        try:
            async with aiofiles.open(signaturefile, "rb") as tmpfile:
                data = await tmpfile.read()
                find_all_signatures = await self._store_signature(
                    digest=digest, signature=data
                )
                return find_all_signatures
        finally:
            signaturefile.unlink(missing_ok=True)

    async def atomicverify(
        self, *, digest: FormattedSHA256, image_name: ImageName
    ) -> List[AtomicSignerVerify]:
        # pylint: disable=too-many-locals
        """Verifies atomic signatures."""
        result = []

        # Retrieve all corresponding signatures from all trust store locations ...
        signatures_all = []
        for location in self.locations:
            signatures = await self._find_all_signatures(
                digest=digest, location=location
            )
            signatures_all.extend(signatures)

        # Write the data and signature to temporary files and invoke GnuPG to verify they match ...
        for signature in signatures_all:
            async with aiotempfile(mode="w+b") as signaturefile:
                datafile = Path(f"{signaturefile.name}.atomic_signature")

                # Write the signature to a temporary file ...
                await signaturefile.write(signature.signature)
                await signaturefile.flush()

                # Note: These options deviate from GPGSigner.verify()
                args = [
                    "gpg",
                    "--batch",
                    "--decrypt",
                    "--homedir",
                    str(self.homedir),
                    "--no-options",
                    "--no-emit-version",
                    "--no-tty",
                    "--output",
                    str(datafile),
                    "--status-fd",
                    "2",
                    signaturefile.name,
                ]

                execute_command = await self._exeute_command(args=args)

                # Retrieve the data file and cleanup ...
                try:
                    async with aiofiles.open(datafile, "rb") as tmpfile:
                        data = await tmpfile.read()
                        atomic_signature = AtomicSignature(data)
                finally:
                    datafile.unlink(missing_ok=True)

                # Verify both types of signatures ...
                status_atomic = await self._verify_atomic_signature(
                    atomic_signature=atomic_signature,
                    digest=digest,
                    image_name=image_name,
                )
                status_gpg = await GPGSigner._parse_output(
                    output=execute_command.stderr
                )

                # Assign metadata ...
                signer_long = signer_short = "Signature parsing failed!"
                try:
                    signer_short = (
                        f"keyid={status_gpg.key_id} status={status_gpg.status.value}"
                    )
                    signer_long = "\n".join(
                        [
                            f"{''.ljust(8)}Signature made {status_gpg.timestamp} using key ID {status_gpg.key_id}",
                            "".ljust(12) + status_gpg.username,
                        ]
                    )
                except:  # pylint: disable=bare-except
                    ...

                result.append(
                    AtomicSignerVerify(
                        atomic_signature=atomic_signature,
                        fingerprint=status_gpg.fingerprint,
                        key_id=status_gpg.key_id,
                        signer_long=signer_long,
                        signer_short=signer_short,
                        status_atomic=status_atomic,
                        status_gpg=status_gpg.status.value,
                        timestamp=status_gpg.timestamp,
                        trust=status_gpg.trust.value,
                        type="atomicsigner",
                        username=status_gpg.username,
                        valid=(
                            status_atomic is None
                            and status_gpg.status == GPGStatus.VALIDSIG
                            and status_gpg.trust in [GPGTrust.FULLY, GPGTrust.ULTIMATE]
                        ),
                    )
                )

        return result



#!/usr/bin/env python

# pylint: disable=protected-access

"""Operator release helpers."""

import gzip
import logging
import sqlite3
import re
import tarfile

from pathlib import Path
from typing import Dict, List, NamedTuple, Optional

from aiotempfile.aiotempfile import open as aiotempfile
from docker_registry_client_async import (
    FormattedSHA256,
    ImageName,
)
from docker_sign_verify import RegistryV2, RegistryV2ManifestList
from docker_sign_verify.utils import be_kind_rewind

from .atomicsigner import AtomicSignerVerify
from .singleassignment import SingleAssignment
from .specs import OperatorReleaseSpecs
from .utils import (
    copy_image,
    read_from_tar,
    retrieve_and_verify_release_metadata,
    TypingRegexSubstitution,
)

LOGGER = logging.getLogger(__name__)


class TypingOperatorMetadata(NamedTuple):
    # pylint: disable=missing-class-docstring
    bundle: str
    channel: str
    images: List[ImageName]
    package: str


class TypingGetReleaseMetadata(NamedTuple):
    # pylint: disable=missing-class-docstring
    index_database: bytes
    index_name: ImageName
    manifest_digest: FormattedSHA256
    operators: List[TypingOperatorMetadata]
    signature_stores: List[str]
    signatures: List[AtomicSignerVerify]
    signing_keys: List[str]


class TypingSearchLayer(NamedTuple):
    # pylint: disable=missing-class-docstring
    index_database: Optional[bytes]


async def _get_index_db(
    *, tar_file, tarinfo: tarfile.TarInfo, path: Path
) -> Optional[bytes]:
    """
    Retrieves the index database from a given tarinfo, if available.

    Args:
        tar_file: The tar_file of the layer being processed.
        tarinfo: The tarinfo from which to retrieve the index database.
        path: Relative path of the tarinfo within the tar file.

    Returns:
        The index database, or None.
    """
    if path.name == OperatorReleaseSpecs.INDEX_DATABASE_NAME:
        LOGGER.debug("Found index database: %s", path)
        result = await read_from_tar(tar_file, tarinfo)
        return result


async def _search_layer(
    *,
    index_name: ImageName,
    layer: FormattedSHA256,
    registry_v2: RegistryV2,
) -> TypingSearchLayer:
    """
    Searches image layers in a given index image for the index database.

    Args:
        index_name: The name of the index image.
        layer: The image layer to be searched.
        registry_v2: The underlying registry v2 image source to use to retrieve the index database.

    Returns:
        index_database: The index database, or None.
    """
    LOGGER.debug("Extracting from layer : %s", layer)

    index_database = SingleAssignment("index_database")
    async with aiotempfile(mode="w+b") as file:
        await registry_v2.docker_registry_client_async.get_blob_to_disk(
            digest=layer, file=file, image_name=index_name
        )
        await be_kind_rewind(file)

        with gzip.GzipFile(filename=file.name) as gzip_file_in:
            with tarfile.open(fileobj=gzip_file_in) as tar_file_in:
                for tarinfo in tar_file_in:
                    path = Path(tarinfo.name)

                    if not str(path).startswith(
                        OperatorReleaseSpecs.DATABASE_PATH_PREFIX
                    ):
                        continue
                    tmp = await _get_index_db(
                        tar_file=tar_file_in, tarinfo=tarinfo, path=path
                    )
                    if tmp:
                        index_database.set(tmp)
    return TypingSearchLayer(index_database=index_database.get())


async def get_release_metadata(
    *,
    index_name: ImageName,
    package_channel: Dict[str, str] = None,
    registry_v2: RegistryV2,
    signature_stores: List[str] = None,
    signing_keys: List[str] = None,
    verify: bool = True,
) -> TypingGetReleaseMetadata:
    # pylint: disable=too-many-branches,too-many-locals,too-many-statements
    """
    Retrieves all metadata for a given package name(s).

    Args:
        index_name: The operator release image for which to retrieve the metadata.
        package_channel: Mapping of package names to content channels. Providing 'None' as the channel name will use the
                         default channel from the index database.
        registry_v2: The Registry V2 image source to use to connect.
        signature_stores: A list of signature store uri overrides.
        signing_keys: A list of armored GnuPG trust store overrides.
        verify: If True, the atomic signature will be retrieved and validated.

    Returns:
        dict:
            index_database: The index database containing all operator metadata.
            operators: A list of process operator metadata.
            signature_stores: A list of signature store uris.
            signing_keys: A list of armored GnuPG trust stores.
    """
    LOGGER.debug("Source index name: %s", index_name)

    # Retrieve the manifest ...
    manifest = await registry_v2.get_manifest(image_name=index_name)
    if isinstance(manifest, RegistryV2ManifestList):
        LOGGER.debug("Manifest list digest: %s", manifest.get_digest())
        manifest = await registry_v2._get_manifest_from_manifest_list(
            image_name=index_name, manifest_list=manifest
        )
    manifest_digest = manifest.get_digest()
    LOGGER.debug("Source index manifest digest: %s", manifest_digest)

    # Search through all layers in reverse order, looking for index.db under a given prefix ...
    index_database = SingleAssignment("index_database")
    for layer in manifest.get_layers():
        tmp = await _search_layer(
            layer=FormattedSHA256.parse(layer),
            registry_v2=registry_v2,
            index_name=index_name,
        )
        if tmp.index_database:
            index_database.set(tmp.index_database)
    index_database = index_database.get()
    if not index_database:
        raise Exception("Unable to locate index database!")

    # Extract the index to a temporary location for processing ...
    async with aiotempfile(mode="w+b") as file:
        await file.write(index_database)
        await file.flush()
        connection = sqlite3.connect(file._file.name)
        cursor = connection.cursor()

        # Start by retrieving all packages and default channels as the initial filter set ...
        rows = cursor.execute("SELECT name, default_channel FROM package")
        package_channel_filtered = {row[0]: row[1] for row in rows}
        LOGGER.debug(
            "Discovered %d packages in index database.", len(package_channel_filtered)
        )

        # We must be able to retrieve information from the index database
        if not package_channel_filtered:
            raise Exception("Unable retrieve package metadata from index database!")

        # If the user provided a filter set, use that instead, but populate missing channels with the defaults ...
        if package_channel:
            # All user provided package names must exist within the index database
            for package in package_channel.keys():
                if package not in package_channel_filtered:
                    raise Exception(f"Unable to locate package: {package}")
            for package in package_channel.keys():
                if package_channel[package] is None:
                    package_channel[package] = package_channel_filtered[package]
            package_channel_filtered = package_channel
        LOGGER.debug("Processing %d package(s).", len(package_channel_filtered))

        # Using the filtered packages and derived channels retrieve the bundles ...
        package_bundle = {}
        for package, channel in package_channel_filtered.items():
            rows = cursor.execute(
                "SELECT head_operatorbundle_name FROM channel WHERE name=:channel and package_name=:package",
                {"channel": channel, "package": package},
            ).fetchall()
            if len(rows) != 1:
                raise Exception(
                    f"Unexpected number of rows returned ({len(rows)}) for package, channel: {package}, {channel}"
                )
            package_bundle[package] = rows[0][0]
        LOGGER.debug("Processing %d bundle(s).", len(package_bundle))

        # Using the packages and bundles retrieved all images ...
        package_images = {}
        for package, bundle in package_bundle.items():
            rows = cursor.execute(
                "SELECT image FROM related_image where operatorbundle_name=:bundle",
                {"bundle": bundle},
            ).fetchall()
            if len(rows) < 1:
                raise Exception(f"No images found relating to bundle: {bundle}s")
            package_images[package] = [ImageName.parse(row[0]) for row in rows]
        LOGGER.debug(
            "Discovered %d related images.",
            sum([len(value) for key, value in package_images.items()]),
        )

    signatures = []
    # TODO: Do we want to default signing keys and signature stores?
    if verify:
        if not signature_stores or not signing_keys:
            raise RuntimeError(
                "Unable to verify release authenticity; signing_key or signature_store not provided!"
            )

        signatures = await retrieve_and_verify_release_metadata(
            digest=manifest_digest,
            image_name=index_name,
            keys=signing_keys,
            locations=signature_stores,
            registry_v2=registry_v2,
        )
    else:
        LOGGER.debug("Skipping source release authenticity verification!")

    result = TypingGetReleaseMetadata(
        index_database=index_database,
        index_name=index_name,
        manifest_digest=manifest_digest,
        operators=[
            TypingOperatorMetadata(
                bundle=package_bundle[package],
                channel=package_channel_filtered[package],
                images=package_images[package],
                package=package,
            )
            for package in package_channel_filtered.keys()
        ],
        signature_stores=signature_stores,
        signatures=signatures,
        signing_keys=signing_keys,
    )

    return result


async def log_release_metadata(
    *,
    index_name: ImageName,
    release_metadata: TypingGetReleaseMetadata,
    sort_metadata: bool = False,
):
    """
    Appends metadata for a given release to the log

    Args:
        index_name: The operator release image for which to retrieve the metadata.
        release_metadata: The metadata for the release to be logged.
        sort_metadata: If True, the metadata keys will be sorted.
    """
    LOGGER.info(index_name)
    operators = release_metadata.operators
    if sort_metadata:
        operators = sorted(release_metadata.operators, key=lambda x: x.package)
    for operator in operators:
        LOGGER.info(
            "  package: %s  bundle: %s  channel: %s  images: %d",
            operator.package,
            operator.bundle,
            operator.channel,
            len(operator.images),
        )

        images = operator.images
        if sort_metadata:
            images = sorted(images)
        for image in images:
            LOGGER.info("    %s", image)
        LOGGER.info("")

    LOGGER.info("  signatures:")
    signatures = release_metadata.signatures
    if sort_metadata:
        signatures = sorted(signatures, key=lambda item: item.timestamp, reverse=True)
    for signature in signatures:
        LOGGER.info(
            "    release name    : %s",
            signature.atomic_signature.get_docker_reference(),
        )
        LOGGER.info(
            "    manifest digest : %s",
            signature.atomic_signature.get_docker_manifest_digest(),
        )
        LOGGER.info("    fingerprint     : %s", signature.key_id)
        LOGGER.info("    username        : %s", signature.username)
        LOGGER.info("    timestamp       : %s", signature.timestamp)
        LOGGER.info("")


async def put_release(
    *,
    index_name: ImageName,
    registry_v2: RegistryV2,
    release_metadata: TypingGetReleaseMetadata,
    verify: bool = True,
):
    """
    Mirrors an openshift release.

    Args:
        index_name: The operator release image to which to store the metadata.
        registry_v2: The Registry V2 image source to use to connect.
        release_metadata: The metadata for the release to be mirrored.
        verify: If True, the atomic signature will be retrieved and validated.
    """
    LOGGER.debug("Destination index image name: %s", index_name)

    LOGGER.debug(
        "Replicating %d operators with %d images.",
        len(release_metadata.operators),
        sum([len(operator.images) for operator in release_metadata.operators]),
    )

    if verify:
        if not release_metadata.signature_stores or not release_metadata.signing_keys:
            raise RuntimeError(
                "Unable to verify release authenticity; signing_key or signature_store not provided!"
            )
        await retrieve_and_verify_release_metadata(
            digest=release_metadata.manifest_digest,
            image_name=index_name,
            keys=release_metadata.signing_keys,
            locations=release_metadata.signature_stores,
            registry_v2=registry_v2,
        )
    else:
        LOGGER.debug("Skipping source release authenticity verification!")

    # Replicate all operator images images ...
    for operator in release_metadata.operators:
        for image_name_src in operator.images:
            image_name_dest = image_name_src.clone()
            image_name_dest.endpoint = index_name.endpoint
            await copy_image(
                image_name_dest=image_name_dest,
                image_name_src=image_name_src,
                registry_v2=registry_v2,
            )

    # Replicate the index ...
    await copy_image(
        image_name_dest=index_name,
        image_name_src=release_metadata.index_name,
        registry_v2=registry_v2,
    )


async def translate_release_metadata(
    *,
    regex_substitutions: List[TypingRegexSubstitution],
    release_metadata: TypingGetReleaseMetadata,
    signature_stores: List[str] = None,
    signing_keys: List[str] = None,
) -> TypingGetReleaseMetadata:
    """
    Translates metadata for a given release to support hop 2-n mirroring.

    Args:
        regex_substitutions: Regular expression substitutions to be applied to source uris.
        release_metadata: The metadata for the release to be translated.
        signature_stores: A list of signature store uri overrides.
        signing_keys: A list of armored GnuPG trust store overrides.

    Returns:
        The translated metadata for a given release.
    """
    if signature_stores is None:
        signature_stores = release_metadata.signature_stores
    if signing_keys is None:
        signing_keys = release_metadata.signing_keys

    operators = release_metadata.operators
    for regex_substitution in regex_substitutions:
        pattern = re.compile(regex_substitution.pattern)
        operators = [
            TypingOperatorMetadata(
                bundle=operator.bundle,
                channel=operator.channel,
                images=[
                    ImageName.parse(
                        pattern.sub(regex_substitution.replacement, str(image))
                    )
                    for image in operator.images
                ],
                package=operator.package,
            )
            for operator in release_metadata.operators
        ]

    return TypingGetReleaseMetadata(
        index_database=release_metadata.index_database,
        index_name=release_metadata.index_name,
        manifest_digest=release_metadata.manifest_digest,
        operators=operators,
        signatures=release_metadata.signatures,
        signature_stores=signature_stores,
        signing_keys=signing_keys,
    )



#!/usr/bin/env python

import os
import re

from setuptools import setup, find_packages


def find_version(*segments):
    root = os.path.abspath(os.path.dirname(__file__))
    abspath = os.path.join(root, *segments)
    with open(abspath, "r") as file:
        content = file.read()
    match = re.search(r"^__version__ = ['\"]([^'\"]+)['\"]", content, re.MULTILINE)
    if match:
        return match.group(1)
    raise RuntimeError("Unable to find version string!")


setup(
    author="Richard Davis",
    author_email="crashvb@gmail.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    description="A utility that can be used to mirror OpenShift releases between docker registries.",
    entry_points="""
        [console_scripts]
        atomic=oc_mirror.scripts.atomic:cli
        oc-mirror=oc_mirror.scripts.oc_mirror:cli
        op-mirror=oc_mirror.scripts.op_mirror:cli
    """,
    extras_require={
        "dev": [
            "black",
            "pylint",
            "pyopenssl",
            "pytest",
            "pytest-asyncio",
            "pytest-docker-apache-fixtures",
            "pytest-docker-registry-fixtures",
            "pytest-docker-squid-fixtures>=0.1.2",
            "pytest_gnupg_fixtures",
            "twine",
            "wheel",
        ]
    },
    include_package_data=True,
    install_requires=[
        "aiofiles",
        "click",
        "docker-registry-client-async>=0.2.1",
        "docker-sign-verify>=2.0.2",
        "pretty-bad-protocol>=3.1.1",
        "pyyaml",
    ],
    keywords="integrity mirror oc oc-mirror openshift sign signatures verify",
    license="GNU General Public License v3.0",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    name="oc_mirror",
    packages=find_packages(),
    project_urls={
        "Bug Reports": "https://github.com/crashvb/oc-mirror/issues",
        "Source": "https://github.com/crashvb/oc-mirror",
    },
    tests_require=[
        "pyopenssl",
        "pytest",
        "pytest-asyncio",
        "pytest-docker-apache-fixtures",
        "pytest-docker-registry-fixtures",
        "pytest-docker-squid-fixtures",
        "pytest_gnupg_fixtures",
    ],
    test_suite="tests",
    url="https://pypi.org/project/oc-mirror/",
    version=find_version("oc_mirror", "__init__.py"),
)



#!/usr/bin/env python

"""
Abstraction of an openshift atomic signature, as defined in:

https://github.com/openshift/containers-image/blob/master/docs/atomic-signature.md
"""

import logging


from docker_registry_client_async import FormattedSHA256, ImageName, JsonBytes

LOGGER = logging.getLogger(__name__)


class AtomicSignature(JsonBytes):
    """
    OpenShift atomic signature.
    """

    TYPE = "atomic container signature"

    @staticmethod
    def minimal(
        *,
        docker_manifest_digest: FormattedSHA256,
        docker_reference: str,
        _type: str = TYPE,
    ) -> "AtomicSignature":
        """
        Creates a minimalistic atomic signature.

        Args:
            docker_manifest_digest: The docker manifest digest.
            docker_reference: The docker reference.
            _type: The signature type.

        Returns:
            The corresponding atomic signature.
        """
        atomic_signature = (
            f'{{"critical":{{"image":{{"docker-manifest-digest": "{docker_manifest_digest}"}},"type": "{_type}",'
            f'"identity": {{"docker-reference": "{docker_reference}"}}}}}}'.encode(
                encoding="utf-8"
            )
        )
        return AtomicSignature(atomic_signature)

    def get_docker_manifest_digest(self) -> FormattedSHA256:
        """
        Retrieves the docker manifest digest.

        Returns:
            The docker manifest digest.
        """
        return FormattedSHA256.parse(
            self.get_json()["critical"]["image"]["docker-manifest-digest"]
        )

    def get_docker_reference(self) -> ImageName:
        """
        Retrieves the docker reference.

        Returns:
            The docker reference.
        """
        return ImageName.parse(
            self.get_json()["critical"]["identity"]["docker-reference"]
        )

    def get_type(self) -> str:
        """
        Retrieves the signature type.

        Returns:
            The signature type.
        """
        return self.get_json()["critical"]["type"]



#!/usr/bin/env python

# pylint: disable=protected-access,redefined-outer-name

"""GPGSigner tests."""

import asyncio
import logging

from ssl import create_default_context
from time import time
from typing import Generator

import pytest

from aiohttp.helpers import BasicAuth
from docker_registry_client_async import (
    DockerRegistryClientAsync,
    FormattedSHA256,
    ImageName,
)
from docker_sign_verify import GPGTrust
from pytest_docker_apache_fixtures import ApacheSecure
from pytest_gnupg_fixtures import GnuPGKeypair
from pytest_docker_squid_fixtures import SquidSecure

from oc_mirror import AtomicSigner

pytestmark = [pytest.mark.asyncio]

LOGGER = logging.getLogger(__name__)

# Bug Fix: https://github.com/crashvb/docker-registry-client-async/issues/24
#
# Right now this is known to leave a nasty "Fatal error on SSL transport" error
# at the end of the test execution; however, without this we cannot test using
# a TLS-in-TLS proxy ...
setattr(asyncio.sslproto._SSLProtocolTransport, "_start_tls_compatible", True)


@pytest.fixture
async def docker_registry_client_async_proxy(
    apache_secure: ApacheSecure, squid_secure: SquidSecure
) -> Generator[DockerRegistryClientAsync, None, None]:
    """Provides a DockerRegistryClientAsync instance."""
    ssl_context = create_default_context(
        cadata=squid_secure.certs.ca_certificate.read_text("utf-8")
        + apache_secure.certs.ca_certificate.read_text("utf-8")
    )
    # Do not use caching; get a new instance for each test
    async with DockerRegistryClientAsync(
        ssl=ssl_context
    ) as docker_registry_client_async:
        credentials = apache_secure.auth_header["Authorization"].split()[1]
        for name in [
            apache_secure.endpoint,
            apache_secure.endpoint_name,
        ]:
            await docker_registry_client_async.add_credentials(name, credentials)
        docker_registry_client_async.proxies[
            "https"
        ] = f"https://{squid_secure.endpoint}/"
        docker_registry_client_async.proxy_auth = BasicAuth(
            login=squid_secure.username, password=squid_secure.password
        )

        yield docker_registry_client_async


@pytest.fixture()
def signaturestore_proxy(apache_secure: ApacheSecure) -> str:
    """Provides a modifiable signature store location."""
    return f"https://{apache_secure.endpoint_name}"


@pytest.fixture()
def atomicsigner_proxy(
    docker_registry_client_async_proxy: DockerRegistryClientAsync,
    gnupg_keypair: GnuPGKeypair,
    signaturestore_proxy: str,
) -> AtomicSigner:
    """Provides AtomicSigner instances."""
    return AtomicSigner(
        docker_registry_client_async=docker_registry_client_async_proxy,
        homedir=gnupg_keypair.gnupg_home,
        keyid=gnupg_keypair.fingerprints[1],
        locations=[signaturestore_proxy],
        passphrase=gnupg_keypair.passphrase,
    )


# TODO: Add tests for protected methods ...


async def test_simple(atomicsigner_proxy: AtomicSigner, gnupg_keypair: GnuPGKeypair):
    """Test configuration signing and verification using GPG."""
    digest = FormattedSHA256.calculate(f"TEST DATA: {time()}".encode(encoding="utf-8"))
    image_name = ImageName.parse(f"foo/bar:{time()}")
    LOGGER.debug("Test Data:\n  digest     : %s\n  image name : %s", digest, image_name)

    # Generate a signature for the test data ...
    find_all_signatures = await atomicsigner_proxy.atomicsign(
        digest=digest, image_name=image_name
    )
    assert find_all_signatures
    assert find_all_signatures.index
    assert find_all_signatures.signature
    assert find_all_signatures.url
    assert all(
        (
            value in find_all_signatures.url
            for value in ["https", digest.replace(":", "="), "signature-"]
        )
    )

    # Verify the generated signature against the test data ...
    results = await atomicsigner_proxy.atomicverify(
        digest=digest, image_name=image_name
    )
    assert results
    for result in results:
        assert result.fingerprint == atomicsigner_proxy.keyid
        assert atomicsigner_proxy.keyid.endswith(result.key_id)
        assert "failed" not in result.signer_long
        assert "failed" not in result.signer_short
        assert result.status_atomic is None
        assert result.status_gpg == "signature valid"
        assert result.timestamp
        assert result.trust == GPGTrust.ULTIMATE.value
        assert result.type == "atomicsigner"
        assert result.username == gnupg_keypair.uids[0]
        assert result.valid


@pytest.mark.skip("Test scenario not defined.")
async def test_bad_data(atomicsigner_proxy: AtomicSigner, gnupg_keypair: GnuPGKeypair):
    """Test configuration signing and verification using GPG with bad data."""
    digest = FormattedSHA256.calculate(f"TEST DATA: {time()}".encode(encoding="utf-8"))
    image_name = ImageName.parse(f"foo/bar:{time()}")
    LOGGER.debug("Test Data:\n  digest     : %s\n  image name : %s", digest, image_name)

    # Generate a signature for the test data ...
    find_all_signatures = await atomicsigner_proxy.atomicsign(
        digest=digest, image_name=image_name
    )
    assert find_all_signatures
    assert find_all_signatures.index
    assert find_all_signatures.signature
    assert find_all_signatures.url
    assert all(
        (
            value in find_all_signatures.url
            for value in ["https", digest.replace(":", "="), "signature-"]
        )
    )

    # TODO: How to tamper with the data (e.g. produce and stage a signature with a mismatched digest)?

    # Verify the generated signature against the test data ...
    results = await atomicsigner_proxy.atomicverify(
        digest=digest, image_name=image_name
    )
    assert results
    for result in results:
        assert result.fingerprint is None
        assert atomicsigner_proxy.keyid.endswith(result.key_id)
        assert "failed" not in result.signer_long
        assert "failed" not in result.signer_short
        assert result.status_atomic is not None
        assert result.status_gpg != "signature valid"
        assert result.timestamp is None
        assert result.trust == GPGTrust.UNDEFINED.value
        assert result.type == "gpg"
        assert result.username == gnupg_keypair.uids[0]
        assert not result.valid



#!/usr/bin/env python

# pylint: disable=protected-access,redefined-outer-name

"""Operator release tests."""

import asyncio
import logging

from typing import Dict, Generator, List, NamedTuple, Optional

import pytest

from docker_registry_client_async import FormattedSHA256, ImageName
from docker_sign_verify import NoSignatureError, RegistryV2
from _pytest.logging import LogCaptureFixture
from pytest_docker_registry_fixtures import DockerRegistrySecure

from oc_mirror.oprelease import (
    get_release_metadata,
    log_release_metadata,
    put_release,
    translate_release_metadata,
    TypingRegexSubstitution,
)

from .testutils import equal_if_unqualified, get_test_data, needs_credentials

pytestmark = [pytest.mark.asyncio]

LOGGER = logging.getLogger(__name__)

# Bug Fix: https://github.com/crashvb/docker-registry-client-async/issues/24
#
# Right now this is known to leave a nasty "Fatal error on SSL transport" error
# at the end of the test execution; however, without this we cannot test using
# a TLS-in-TLS proxy ...
setattr(asyncio.sslproto._SSLProtocolTransport, "_start_tls_compatible", True)


class TypingGetTestDataLocal(NamedTuple):
    # pylint: disable=missing-class-docstring
    index_name: ImageName
    package_channel: Dict[str, Optional[str]]
    signature_stores: List[str]
    signing_keys: List[str]


def get_release_data() -> Generator[TypingGetTestDataLocal, None, None]:
    """Dynamically initializes test data for a local mutable registry."""
    dataset = [
        TypingGetTestDataLocal(
            index_name=ImageName.parse(
                "registry.redhat.io/redhat/redhat-operator-index:v4.8"
            ),
            package_channel={"ocs-operator": None},
            signature_stores=[
                "https://mirror.openshift.com/pub/openshift-v4/signatures/openshift/release"
            ],
            signing_keys=[],
        ),
        TypingGetTestDataLocal(
            index_name=ImageName.parse(
                "registry.redhat.io/redhat/redhat-operator-index:v4.8"
            ),
            package_channel={"ocs-operator": "eus-4.8"},
            signature_stores=[
                "https://mirror.openshift.com/pub/openshift-v4/signatures/openshift/release"
            ],
            signing_keys=[],
        ),
    ]
    for data in dataset:
        yield data


@pytest.fixture(params=get_release_data())
def known_good_release(request) -> TypingGetTestDataLocal:
    """Provides 'known good' metadata for a local release that can be modified."""
    signing_key = get_test_data(
        request, __name__, "567e347ad0044ade55ba8a5f199e2f91fd431d51.gnupg"
    )
    return TypingGetTestDataLocal(
        index_name=request.param.index_name,
        package_channel=request.param.package_channel,
        signature_stores=request.param.signature_stores,
        signing_keys=[signing_key],
    )


@pytest.mark.online
@needs_credentials("registry.redhat.io")
async def test_get_release_metadata(
    known_good_release: TypingGetTestDataLocal,
    registry_v2_proxy: RegistryV2,
):
    """Tests release metadata retrieval from a remote registry."""
    logging.getLogger("gnupg").setLevel(logging.FATAL)

    # An exception should be raised if the image configuration is not signed
    with pytest.raises(NoSignatureError) as exception:
        await get_release_metadata(
            registry_v2=registry_v2_proxy,
            index_name=known_good_release.index_name,
            package_channel=known_good_release.package_channel,
            signature_stores=known_good_release.signature_stores,
            signing_keys=known_good_release.signing_keys,
        )
    assert str(exception.value) == "Unable to locate a valid signature!"

    # Retrieve the release metadata (which is unsigned from the vendor / community =/ ) ...
    result = await get_release_metadata(
        registry_v2=registry_v2_proxy,
        index_name=known_good_release.index_name,
        package_channel=known_good_release.package_channel,
        signature_stores=known_good_release.signature_stores,
        signing_keys=known_good_release.signing_keys,
        verify=False,
    )

    assert result.index_database is not None
    assert result.manifest_digest
    assert result.operators
    assert result.signature_stores
    assert not result.signatures
    assert result.signing_keys
    assert len(result.operators) == len(known_good_release.package_channel.keys())
    for package in known_good_release.package_channel.keys():
        operator = [
            operator for operator in result.operators if operator.package == package
        ][0]
        assert operator
        assert operator.bundle
        if known_good_release.package_channel[package] is None:
            assert operator.channel is not None
        else:
            assert operator.channel == known_good_release.package_channel[package]
        assert operator.images


@pytest.mark.online
@pytest.mark.parametrize(
    "release,package_channel,bundle_image,bundle_name,related_image",
    [
        (
            "registry.redhat.io/redhat/redhat-operator-index:v4.8",
            {"ocs-operator": "eus-4.8"},
            "registry.redhat.io/ocs4/ocs-operator-bundle@sha256:",
            "ocs-operator.v4.8.8",
            "registry.redhat.io/rhceph/rhceph-4-rhel8@sha256:",
        ),
    ],
)
@needs_credentials("registry.redhat.io")
async def test_log_release_metadata(
    bundle_image: str,
    bundle_name: str,
    caplog: LogCaptureFixture,
    package_channel: Dict[str, str],
    registry_v2_proxy: RegistryV2,
    related_image: str,
    release: str,
):
    # pylint: disable=too-many-arguments
    """Tests logging of release metadata."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    logging.getLogger("gnupg").setLevel(logging.FATAL)

    # Retrieve the release metadata ...
    image_name = ImageName.parse(release)
    result = await get_release_metadata(
        registry_v2=registry_v2_proxy,
        index_name=image_name,
        package_channel=package_channel,
        verify=False,
    )
    assert result

    for sort_metadata in [True, False]:
        await log_release_metadata(
            index_name=image_name, release_metadata=result, sort_metadata=sort_metadata
        )
        assert bundle_image in caplog.text
        assert bundle_name in caplog.text
        assert str(image_name) in caplog.text
        for key in package_channel.keys():
            assert key in caplog.text
        assert related_image in caplog.text


@pytest.mark.online_modification
@needs_credentials("registry.redhat.io")
async def test_put_release_from_internet(
    docker_registry_secure: DockerRegistrySecure,
    known_good_release: TypingGetTestDataLocal,
    registry_v2_proxy: RegistryV2,
):
    """Tests release replication to a local registry."""
    logging.getLogger("gnupg").setLevel(logging.FATAL)

    # Retrieve the release metadata ...
    release_metadata_src = await get_release_metadata(
        registry_v2=registry_v2_proxy,
        index_name=known_good_release.index_name,
        package_channel=known_good_release.package_channel,
        verify=False,
    )

    # Replicate the release ...
    image_name_dest = known_good_release.index_name.clone()
    image_name_dest.endpoint = docker_registry_secure.endpoint_name
    await put_release(
        index_name=image_name_dest,
        registry_v2=registry_v2_proxy,
        release_metadata=release_metadata_src,
        verify=False,
    )

    # Retrieve the release metadata (again) ...
    release_metadata_dest = await get_release_metadata(
        registry_v2=registry_v2_proxy,
        index_name=image_name_dest,
        package_channel=known_good_release.package_channel,
        verify=False,
    )

    # Release metadata should have the index database ...
    assert FormattedSHA256.calculate(
        release_metadata_dest.index_database
    ) == FormattedSHA256.calculate(release_metadata_src.index_database)

    # Release metadata should have the same index name ...
    assert equal_if_unqualified(
        image_name0=release_metadata_src.index_name,
        image_name1=release_metadata_dest.index_name,
    )

    # Release metadata should have the same manifest digest ...
    assert release_metadata_dest.manifest_digest == release_metadata_src.manifest_digest

    # Release metadata should have the same operators listed ...
    assert sorted(
        release_metadata_dest.operators, key=lambda item: item.package
    ) == sorted(release_metadata_src.operators, key=lambda item: item.package)

    # The signature stores should be the same ...
    if release_metadata_src.signature_stores:
        assert (
            release_metadata_dest.signature_stores.sort()
            == release_metadata_src.signature_stores.sort()
        )
    else:
        assert not release_metadata_dest.signature_stores

    # TODO: Do we need to check signatures here?

    # The signing keys should be the same ...
    if release_metadata_src.signing_keys:
        assert (
            release_metadata_dest.signing_keys.sort()
            == release_metadata_src.signing_keys.sort()
        )
    else:
        assert not release_metadata_dest.signing_keys


@pytest.mark.online_modification
@needs_credentials("registry.redhat.io")
async def test_put_release_from_internal(
    docker_registry_secure_list: List[DockerRegistrySecure],
    known_good_release: TypingGetTestDataLocal,
    registry_v2_list_proxy: RegistryV2,
):
    # pylint: disable=too-many-locals
    """Tests release replication to a local registry."""
    logging.getLogger("gnupg").setLevel(logging.FATAL)

    # Retrieve the release metadata (hop 0)...
    image_name0 = known_good_release.index_name.clone()
    release_metadata0 = await get_release_metadata(
        index_name=image_name0,
        package_channel=known_good_release.package_channel,
        registry_v2=registry_v2_list_proxy,
        verify=False,
    )

    # Replicate the release (hop 1)...
    image_name1 = image_name0.clone()
    image_name1.endpoint = docker_registry_secure_list[0].endpoint_name
    await put_release(
        index_name=image_name1,
        registry_v2=registry_v2_list_proxy,
        release_metadata=release_metadata0,
        verify=False,
    )

    # Retrieve the release metadata (hop 1) ...
    release_metadata1 = await get_release_metadata(
        index_name=image_name1,
        package_channel=known_good_release.package_channel,
        registry_v2=registry_v2_list_proxy,
        verify=False,
    )

    # Translate to the second registry ...
    regex_substitutions = [
        TypingRegexSubstitution(
            pattern=r"quay\.io",
            replacement=docker_registry_secure_list[0].endpoint_name,
        ),
        TypingRegexSubstitution(
            pattern=r"registry\.redhat\.io",
            replacement=docker_registry_secure_list[0].endpoint_name,
        ),
    ]
    release_metadata1_translated = await translate_release_metadata(
        regex_substitutions=regex_substitutions, release_metadata=release_metadata1
    )

    # TODO: We need to implement "signing" here, and test with verify=True below ...

    # Replicate the release (hop 2) ...
    image_name2 = image_name0.clone()
    image_name2.endpoint = docker_registry_secure_list[1].endpoint_name
    await put_release(
        index_name=image_name2,
        registry_v2=registry_v2_list_proxy,
        release_metadata=release_metadata1_translated,
        verify=False,
    )

    # Retrieve the release metadata (hop 2) ...
    release_metadata2 = await get_release_metadata(
        index_name=image_name2,
        package_channel=known_good_release.package_channel,
        registry_v2=registry_v2_list_proxy,
        verify=False,
    )

    # Release metadata should have the index database ...
    assert FormattedSHA256.calculate(
        release_metadata2.index_database
    ) == FormattedSHA256.calculate(release_metadata0.index_database)

    # Release metadata should have the same index name ...
    assert equal_if_unqualified(
        image_name0=release_metadata0.index_name,
        image_name1=release_metadata2.index_name,
    )

    # Release metadata should have the same manifest digest ...
    assert release_metadata2.manifest_digest == release_metadata0.manifest_digest

    # Release metadata should have the same operators listed ...
    assert sorted(release_metadata2.operators, key=lambda item: item.package) == sorted(
        release_metadata0.operators, key=lambda item: item.package
    )

    # The signature stores should be the same ...
    if release_metadata0.signature_stores:
        assert (
            release_metadata2.signature_stores.sort()
            == release_metadata0.signature_stores.sort()
        )
    else:
        assert not release_metadata2.signature_stores

    # TODO: Do we need to check signatures here?

    # The signing keys should be the same ...
    if release_metadata0.signing_keys:
        assert (
            release_metadata2.signing_keys.sort()
            == release_metadata0.signing_keys.sort()
        )
    else:
        assert not release_metadata2.signing_keys
